/*
The copyright in this software is being made available under this MPEG Reference Software Copyright License. This software may be subject to other third party and contributor rights, including patent rights, and no such rights are granted under this license.

Copyright (c) 2016-2017, Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
*	Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
*	Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
*	Neither the names of Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms nor the names of their contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "CdvaImpl.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <opencv2/opencv.hpp>
#include <algorithm>
#include "CdvsException.h"

#include "RefFeature.h"

#include "evx/base.h"
#include "evx/cabac.h"
#include "evx/bitstream.h"
#include "evx/math.h"
#include "evx/memory.h"

#include "CoordinateCoding.h"
#include "CompressedFeatureList.h"

#include <time.h>


using namespace cv;
using namespace mpeg7cdva;
using namespace mpeg7cdvs;
using namespace std;

#define FOURCC_MP43 0x3334504D		// identify MS codec MP43


///////////////////////////////// class SegmentDescriptor //////////////////////////////////

SegmentDescriptor::SegmentDescriptor()
{
}

SegmentDescriptor::~SegmentDescriptor()
{
}

bool SegmentDescriptor::empty() const
{
	return (keyframes.size() == 0);
}

void SegmentDescriptor::clear()
{
	keyframes.clear();
}


void SegmentDescriptor::setStartTimeMs(unsigned long position_ms)
{
	startTime = position_ms;
}

void SegmentDescriptor::setEndTimeMs(unsigned long position_ms)
{
	endTime = position_ms;
}


unsigned long SegmentDescriptor::getStartTimeMs() const
{
	return startTime;
}

unsigned long SegmentDescriptor::getEndTimeMs() const
{
	return endTime;
}

void SegmentDescriptor::setParity(unsigned int value)
{
	parity = value & 0x0F;
}

unsigned int SegmentDescriptor::getParity() const
{
	return parity;
}

size_t SegmentDescriptor::getSize() const
{
	size_t mysize = 0;

	for (int i=0; i< (int) keyframes.size(); ++i)
		mysize += keyframes[i].buffer.size();

	return mysize;
}

size_t SegmentDescriptor::write(std::ofstream & fout) const
{
	unsigned char tmpbuffer[80];

	if (empty())
		return 0;		// storing an empty shot is useless

	unsigned int flags = 0; // reset flags
	if (startTime < 0xFFFFFF)
		flags |= 0x01;		// shot start time is valid (fits in 24 bits)

	if (endTime < 0xFFFFFF)
		flags |= 0x02;		// shot end time is valid

	size_t mysize = getSize();
	if (mysize < 0xFFFFFF)
		flags |= 0x04;		// shot size is valid


	BitOutputStream writer;
	writer.open(tmpbuffer, 80);
	writer.write(flags, 4);							// write flags (4 bits)
	writer.write(parity, 4);						// write shot parity (4 bits)
	writer.write((unsigned int) startTime, 24);		// write shot time
	writer.write((unsigned int) endTime,   24);		// write frame time

	writer.write((unsigned int) mysize, 24); 		// write total buffer size
	writer.close();									// close bit writer

	fout.write((char*)tmpbuffer, SHOT_HEADER_SIZE);	// write header into a FILE

	if (flags == 0x07)		// write data only if all flags are OK
	{
		for (int i=0; i < (int) keyframes.size(); ++i) {
			fout.write((const char *) keyframes[i].buffer.data(), keyframes[i].buffer.size());	// write descriptor into a FILE
		}
		return SHOT_HEADER_SIZE + mysize;
	}
	else
	{
		cerr << "ERROR; wrong flags writing shot: " << flags << endl;
		return SHOT_HEADER_SIZE;
	}
}

size_t SegmentDescriptor::read(std::ifstream & fin, mpeg7cdvs::CdvsServer * cdvsserver)
{
	unsigned char tmpbuffer[80];
	fin.read((char*)&tmpbuffer, SHOT_HEADER_SIZE);

	if (fin.good())
	{
		BitInputStream reader;
		reader.open(tmpbuffer, 80);

		unsigned int flags = reader.read(4);
		parity = reader.read(4);
		startTime = reader.read(24);
		endTime = reader.read(24);
		unsigned int mysize = reader.read(24);
		reader.close();

		if ((flags & 0x01) == 0)
			startTime = 0xFFFFFF; 	// shot start time is not valid

		if ((flags & 0x02) == 0)
			endTime = 0xFFFFFF; 	// shot end time is not valid

		if ((flags & 0x04) == 0)	// buffer size is not valid
			mysize = 0xFFFFFF;

		if (flags == 0x07)		// read data only if all flags are OK
		{
			Buffer readbuf(mysize);
			fin.read(readbuf.sdata(), mysize);	// read data from file

			unsigned int consumed = 0;

			while (consumed < mysize)
			{
				CdvsDescriptor descr;
				consumed += cdvsserver->decode (descr, readbuf.data() + consumed, readbuf.size() - consumed);
				keyframes.push_back(descr);
			}

			return SHOT_HEADER_SIZE + mysize;
		}

		return SHOT_HEADER_SIZE;
	}

	return 0;
}

size_t SegmentDescriptor::addSegmentToDB(std::ifstream & fin, mpeg7cdvs::CdvsServer * cdvsserver, const std::string & relativepathname, bool optMatch)
{
	unsigned char tmpbuffer[80];
	fin.read((char*)&tmpbuffer, SHOT_HEADER_SIZE);

	if (fin.good())
	{
		BitInputStream reader;
		reader.open(tmpbuffer, 80);

		unsigned int flags = reader.read(4);
		int parity = reader.read(4);
		int startTime = reader.read(24);
		int endTime = reader.read(24);
		unsigned int mysize = reader.read(24);
		reader.close();

		if ((flags & 0x01) == 0)
			startTime = 0xFFFFFF; 	// shot start time is not valid

		if ((flags & 0x02) == 0)
			endTime = 0xFFFFFF; 	// shot end time is not valid

		if ((flags & 0x04) == 0)	// buffer size is not valid
			mysize = 0xFFFFFF;

		if (flags == 0x07)		// read data only if all flags are OK
		{
			Buffer readbuf(mysize);
			fin.read(readbuf.sdata(), mysize);	// read data from file

			unsigned int consumed = 0;


			while (consumed < mysize)
			{
				CdvsDescriptor descr;
				consumed += cdvsserver->decode (descr, readbuf.data() + consumed, readbuf.size() - consumed);
				if (descr.featurelist.nFeatures() > 0)
					cdvsserver->addDescriptorToDB(descr, relativepathname.c_str());	// add image to DB

				if (optMatch) {
					// stop after medoid
					break;
				}
			}

			return SHOT_HEADER_SIZE + mysize;
		}

		return SHOT_HEADER_SIZE;
	}

	return 0;
}

///////////////////////////////// class CompressedSegmentDescriptor /////////////////////////////////////////////

CompressedSegmentDescriptor::CompressedSegmentDescriptor(const mpeg7cdvs::Parameters& p) : params(p)
{
	globalBuf = 0;
	globalBufSz = 0;
	localBuf = 0;
	localBufSz = 0;
	histoBuf = 0;
	histoBufSz = 0;
}

unsigned char* CompressedSegmentDescriptor::allocateLocalBuffer(unsigned long sz) {
	if (localBuf) delete[] localBuf;
	localBuf = new unsigned char[sz];
	localBufSz = sz;

	return localBuf;
}

unsigned char*  CompressedSegmentDescriptor::allocateGlobalBuffer(unsigned long sz) {
	if (globalBuf) delete[] globalBuf;
	globalBuf = new unsigned char[sz];
	globalBufSz = sz;

	return globalBuf;
}

unsigned char*  CompressedSegmentDescriptor::allocateHistoBuffer(unsigned long sz) {
	if (histoBuf) delete[] histoBuf;
	histoBuf = new unsigned char[sz];
	histoBufSz = sz;

	return histoBuf;
}

CompressedSegmentDescriptor::~CompressedSegmentDescriptor()
{
	if (globalBuf) delete [] globalBuf;
	if (localBuf) delete[] localBuf;
	if (histoBuf) delete[] histoBuf;
}


size_t CompressedSegmentDescriptor::write(std::ofstream & fout) const
{
	unsigned char tmpbuffer[SHOT_HEADER_SIZE];

	if (empty())
		return 0;		// storing an empty shot is useless

	unsigned int flags = 0; // reset flags
	if (startTime < 0xFFFFFF)
		flags |= 0x01;		// shot start time is valid (fits in 24 bits)

	if (endTime < 0xFFFFFF)
		flags |= 0x02;		// shot end time is valid

	size_t mysize = getSize();
	if (mysize < 0xFFFFFF)
		flags |= 0x04;		// shot size is valid


	BitOutputStream writer;
	writer.open(tmpbuffer, SHOT_HEADER_SIZE*8);
	writer.write(flags, 4);							// write flags (4 bits)
	writer.write(parity, 4);						// write shot parity (4 bits)
	writer.write((unsigned int)(startTime & 0xFFFFFF), 24);		// write shot time
	writer.write((unsigned int)(endTime & 0xFFFFFF), 24);		// write end time

	// size of global buffer of medoid frame
	unsigned int medoidGlobalBufferSize = 512;
	int nFctPresent = numGlobFctPresent(keyframes[0]);
	medoidGlobalBufferSize += ((keyframes[0].getGlobalHasBitSelection() ? 24 : 32) * nFctPresent);
	medoidGlobalBufferSize += ((keyframes[0].getGlobalHasVariance() ? 32 : 0) * nFctPresent); 
	writer.write(medoidGlobalBufferSize, 24);

	writer.write((unsigned int)globalBufSz, 24); 	// write global buffer size
	writer.write((unsigned int)localBufSz, 24); 	// write local buffer size

	writer.write((unsigned int)globalUncompressedBufSz, 24); 	// write uncompressed global buffer size
	writer.write((unsigned int)localUncompressedBufSz, 24); 	// write uncompressed local buffer size
	
	writer.write((unsigned int)histoBufSz, 24); 	// write histogram buffer size

	writer.write((unsigned char)keyframes[0].getModeID(), 8);
	writer.write((unsigned short)keyframes[0].getOriginalImageXResolution(), 16);
	writer.write((unsigned short)keyframes[0].getOriginalImageYResolution(), 16);
	writer.write((unsigned short)keyframes[0].getHistogramMapSizeX(), 16);
	writer.write((unsigned short)keyframes[0].getHistogramMapSizeY(), 16);

	unsigned char descFlags = 0;
	if (keyframes[0].getGlobalHasBitSelection()) descFlags += 1;
	if (keyframes[0].getGlobalHasVariance()) descFlags += 2;
	if (keyframes[0].getRelevanceBitsPresent()) descFlags += 4;
	writer.write((unsigned char)descFlags, 8);

	writer.close();									// close bit writer

	fout.write((char*)tmpbuffer, SHOT_HEADER_SIZE);	// write header into a FILE

	// write descriptor time map
	unsigned int dtmHeaderBufferSz = 5 + 2*(dtm.getNFrames()-1); // nFrames, nAbsDesc, nDesc per diff frame
	unsigned char* buffer = new unsigned char[dtmHeaderBufferSz];
	writer.open(buffer, dtmHeaderBufferSz * 8);
	writer.write((unsigned char)(dtm.getNFrames()), 8);
	writer.write((unsigned short)(dtm.getNAbsDescr()), 16);

	int nTotalDesc = dtm.getNAbsDescr();
	for (int i = 1; i < dtm.getNFrames(); i++) {
		writer.write((unsigned short)(keyframes[i].getNumberOfLocalDescriptors()), 16);
		nTotalDesc += keyframes[i].getNumberOfLocalDescriptors();
	}

	writer.write((unsigned short)(dtm.getNDescr()), 16);

	writer.close();
	fout.write((char*)buffer, dtmHeaderBufferSz);
	
	delete[] buffer;
	
	/*
	unsigned int dtmBufSz = ceil(dtm.getBitSize() / 8.0);
	fout.write(dtm.getBuffer(), dtmBufSz);
	*/
	buffer = new unsigned char[2*nTotalDesc+2];
	int descIdxBufSz = 0;
	dtm.writeDescIndex(buffer, true, descIdxBufSz);
	fout.write((char*)buffer, descIdxBufSz);
	delete[] buffer;

	
	// write global medoid buffer to file
	buffer = new unsigned char[medoidGlobalBufferSize];
	writer.open(buffer, medoidGlobalBufferSize * 8);
	keyframes[0].scfvSignature.write(writer); 
	writer.close();
	fout.write((char*)buffer, medoidGlobalBufferSize);
	delete[] buffer;


	if (flags == 0x07)		// write data only if all flags are OK
	{
	
		// global buffer
		fout.write((const char *)globalBuf, globalBufSz);

		// local buffer
		fout.write((const char *)localBuf, localBufSz);
		
		// histograms
		fout.write((const char *)histoBuf, histoBufSz);

		return SHOT_HEADER_SIZE + ceil(dtm.getBitSize() / 8.0) + (dtm.getNFrames() * 2 + 3) + medoidGlobalBufferSize + globalBufSz + localBufSz + histoBufSz;
	}
	else
	{
		cerr << "ERROR; wrong flags writing shot: " << flags << endl;
		return SHOT_HEADER_SIZE;
	}
}

size_t CompressedSegmentDescriptor::read(std::ifstream & fin, mpeg7cdvs::CdvsServer * cdvsserver)
{
	unsigned char tmpbuffer[SHOT_HEADER_SIZE];
	fin.read((char*)&tmpbuffer, SHOT_HEADER_SIZE);

	if (fin.good())
	{
		BitInputStream reader;
		reader.open(tmpbuffer, SHOT_HEADER_SIZE * 8);

		unsigned int flags = reader.read(4);
		parity = reader.read(4);
		startTime = reader.read(24);
		endTime = reader.read(24);
		unsigned int _medoidGlobalBufSz = reader.read(24);
		unsigned int _globalBufSz = reader.read(24);
		unsigned int _localBufSz = reader.read(24);

		globalUncompressedBufSz = reader.read(24);
		localUncompressedBufSz = reader.read(24);

		unsigned int _histoBufSz = reader.read(24);
		
		int cdvsDescriptorMode = reader.read(8);
		unsigned int origImgX = reader.read(16);
		unsigned int origImgY = reader.read(16);
		unsigned int histoMapSzX = reader.read(16);
		unsigned int histoMapSzY = reader.read(16);
		unsigned char descFlags = reader.read(8);
		
		bool hasBitSel = (descFlags & 0x01)>0;
		bool hasVar = (descFlags & 0x02)>0;
		bool hasRelB = (descFlags & 0x04)>0;

		reader.close();

		if ((flags & 0x01) == 0)
			startTime = 0xFFFFFF; 	// shot start time is not valid

		if ((flags & 0x02) == 0)
			endTime = 0xFFFFFF; 	// shot end time is not valid

		if ((flags & 0x04) == 0) {	// buffer size is not valid
			globalBufSz = 0xFFFFFF;
			localBufSz = 0xFFFFFF;
		}

		if (flags == 0x07)		// read data only if all flags are OK
		{

			// read descriptor time map
			
			// read n frames and n abs descr
			fin.read((char*)&tmpbuffer, 3); 

			BitInputStream reader;
			reader.open(tmpbuffer, 3 * 8);
			unsigned int _nFrames = reader.read(8);			
			unsigned int _nAbsDesc = reader.read(16);
			reader.close();			
			
			// read counts for frames
			unsigned char* countBuf = new unsigned char[(_nFrames)*2];
			fin.read((char*)countBuf, (_nFrames ) * 2);
			reader.open(countBuf, (_nFrames ) * 2 * 8);

			std::vector<unsigned int> nLocalDesc;
			nLocalDesc.push_back(_nAbsDesc);

			int nTotalDesc = _nAbsDesc;
			for (int i = 1; i < _nFrames; i++) {

				unsigned int nld = reader.read(16);
				nLocalDesc.push_back(nld);
				nTotalDesc += nld;
			}

			unsigned int _nDesc = reader.read(16);
			
			reader.close();

			delete[] countBuf;

			// read map
			dtm.init(_nDesc, _nAbsDesc, _nFrames);

			//fin.read(dtm.getBuffer(), ceil(dtm.getBitSize() / 8.0));	// read data from file
			unsigned short descIdxBufSz = 0;
			fin.read((char*)(&descIdxBufSz), 2);
			unsigned char* buffer = new unsigned char[descIdxBufSz];
			fin.read((char*)buffer, descIdxBufSz);
			dtm.parseDescIndex(buffer, nLocalDesc, true);
			delete[] buffer;

			// read global medoid buffer
			buffer = new unsigned char[_medoidGlobalBufSz];
			fin.read((char*)buffer, _medoidGlobalBufSz);

			unsigned char* globalBuf = allocateGlobalBuffer(_globalBufSz);
			unsigned char* localBuf = allocateLocalBuffer(_localBufSz);
			unsigned char* histoBuf = allocateHistoBuffer(_histoBufSz);

			// read compressed buffers
			fin.read((char*)globalBuf, globalBufSz);	// read data from file
			fin.read((char*)localBuf, localBufSz);	// read data from file

			// read histogram buffer
			fin.read((char*)histoBuf, histoBufSz);

			decode((char*)buffer, _medoidGlobalBufSz, nLocalDesc, cdvsDescriptorMode, origImgX, origImgY, histoMapSzX, histoMapSzY, hasBitSel, hasVar, hasRelB);

			delete[] buffer;

			return SHOT_HEADER_SIZE + ceil(dtm.getBitSize() / 8.0) + ((_nFrames ) * 2 + 3) + _medoidGlobalBufSz  + globalBufSz + localBufSz + histoBufSz;
		}

		return SHOT_HEADER_SIZE;
	}

	return 0;
}

void CompressedSegmentDescriptor::decode(char* globalMedBuf, unsigned int globalMedBufSz, std::vector<unsigned int>& nLocalDesc,
	unsigned char cdvsDescriptorMode, unsigned short origImgX, unsigned short origImgY, unsigned short histoMapSzX, unsigned short histoMapSzY,
	bool hasBitSel, bool hasVar, bool hasRelB) {


	// reading CDFVSignature of medoid descriptor (SCVF_med)
	int medoidNumGlobalFctPresent = 0;
	std::vector<int> setIndices;

	for (int q = 0; q < 64; q++) {
		for (int r = 0; r < 8; r++) {
			if ((globalMedBuf[q] & (0x1 << r))>0) {
				medoidNumGlobalFctPresent++;
				setIndices.push_back(q * 8 + r);
			}
		}
	}

	BitInputStream medScvfRead;
	medScvfRead.open((unsigned char*)globalMedBuf, globalMedBufSz);

	CdvsDescriptor medoid;
	medoid.setOriginalImageXResolution(origImgX);
	medoid.setOriginalImageYResolution(origImgY);
	medoid.setHistogramMapSizeX(histoMapSzX);
	medoid.setHistogramMapSizeY(histoMapSzY);
	medoid.setModeID(cdvsDescriptorMode);
	medoid.setGlobalHasBitSelection(hasBitSel);
	medoid.setGlobalHasVariance(hasVar);
	medoid.setRelevanceBitsPresent(hasRelB);
	medoid.setNumberOfLocalDescriptors(dtm.getNAbsDescr());
	medoid.setHistogramCountSize(0); // TODO: check if ok

	medoid.scfvSignature.hasVar(hasVar);
	medoid.scfvSignature.hasBitSelection( hasBitSel );
	
	medoid.scfvSignature.read(medScvfRead);
	medScvfRead.close();

	keyframes.push_back(medoid);

	evx::entropy_coder coder;
	if ((dtm.getNFrames() - 1) > 0) {

		char* uncompressedData = new char[globalUncompressedBufSz];

		evx::bitstream src(globalBufSz * 8);
		evx::bitstream dst(globalUncompressedBufSz * 8);

		src.write_bytes(globalBuf, globalBufSz);

		coder.decode(globalUncompressedBufSz * 8, &src, &dst, true);

		evx::uint32 gubs = (evx::int32)globalUncompressedBufSz;
		dst.read_bytes(uncompressedData, &gubs);

		reconstructGlobalDifferences(uncompressedData, globalUncompressedBufSz, dtm.getNFrames() - 1, hasVar, hasBitSel, nLocalDesc);

		delete[] uncompressedData;

	}


	////////////////// local descriptors ///////////////////

	//mpeg7cdva::CompressedFeatureList cfl;
	mpeg7cdvs::FeatureList fl;

	if (localBufSz > 0) {

		evx::bitstream src;
		evx::bitstream dst;

		dst.clear();
		dst.resize_capacity(localUncompressedBufSz * 8);
		src.clear();
		src.resize_capacity(localBufSz * 8);

		src.write_bytes(localBuf, localBufSz);

		coder.decode(localUncompressedBufSz * 8, &src, &dst);

		unsigned char* featureListBuf = new unsigned char[localUncompressedBufSz];

		// get descriptor list for entire segment

		try {

			evx::uint32 lubs = (evx::int32)localUncompressedBufSz;

			dst.read_bytes(featureListBuf, &lubs);


			fl = readFeatureListFromBinaryAbs(featureListBuf, localUncompressedBufSz, hasRelB, params.numberOfElementGroups);

			delete[]featureListBuf;

		}
		catch (...) {

			delete[]featureListBuf;

		}
	}

	// parse coordinates, this creates the descriptors for each frame
	// with coordinates, but no feature vector
	try {
		BitInputStream bisHisto((unsigned char*)histoBuf, histoBufSz);

		for (int i = 0; i < keyframes.size(); i++){

			CoordinateCoding ccdecoder(params);
			ccdecoder.fromBinary(bisHisto, histoMapSzX, histoMapSzY);
			ccdecoder.generateFeatureList(keyframes[i].featurelist);
			keyframes[i].featurelist.originalWidth = origImgX;
			keyframes[i].featurelist.originalHeight = origImgY;

		}
		bisHisto.close();
	}
	catch (...) {
	}


	// fill per frame descriptor lists
	for (int j = 0; j < keyframes.size(); ++j) {

		int count = 0;

		keyframes[j].featurelist.qdescr_size = params.numberOfElementGroups * 4;

		vector<unsigned short>& descIndex = dtm.getDescIndex(j);

		for (int i = 0; i < descIndex.size(); i++) {
			for (int l = 0; l < keyframes[j].featurelist.qdescr_size; l++) {
				keyframes[j].featurelist.features[i].qdescr[l] = fl.features[descIndex[i]].qdescr[l];
			}
		}

	}

	

}


void CompressedSegmentDescriptor::reconstructGlobalDifferences(char* uncompressedData, unsigned int uncompressedSz, int nrGDesc, bool hasVar, bool globalHasBitSelection, std::vector<unsigned int>& nLocalDesc){

	
	unsigned char* buffer = (unsigned char*)uncompressedData;
	int bufIdx = 0;

	for (int i = 1; i <= nrGDesc; i++){

		CdvsDescriptor desc;
		desc.setGlobalHasVariance(hasVar);
		desc.scfvSignature.hasVar ( hasVar );
		desc.setGlobalHasBitSelection(globalHasBitSelection);
		desc.scfvSignature.hasBitSelection ( globalHasBitSelection );

		desc.setOriginalImageXResolution(keyframes[0].getOriginalImageXResolution());
		desc.setOriginalImageYResolution(keyframes[0].getOriginalImageYResolution());
		desc.setHistogramMapSizeX(keyframes[0].getHistogramMapSizeX());
		desc.setHistogramMapSizeY(keyframes[0].getHistogramMapSizeY());
		desc.setModeID(keyframes[0].getModeID());
		desc.setRelevanceBitsPresent(keyframes[0].getRelevanceBitsPresent());
		desc.setHistogramCountSize(0); // TODO: check if ok

		desc.setNumberOfLocalDescriptors(nLocalDesc[i]);


		int numVisited = 0;
		std::vector<bool> fctPresent;
		unsigned char sh = 0;
		for (int q = 0; q < 512; q++) 
		{
			if (sh == 8) {
				sh = 0;
				bufIdx++;
			}

			if ((buffer[bufIdx] & (1 << sh)) > 0) {
				numVisited++;
				fctPresent.push_back(true);
			}
			else fctPresent.push_back(false);

			sh++;
		}

		bufIdx++;

		for (int q = 0; q < 512; q++) {
			unsigned int value = 0;
			if (fctPresent[q]) {
				unsigned int* uiBuf = (unsigned int*)(buffer+bufIdx);
				value = uiBuf[0];
				bufIdx+=4;

			}
			desc.scfvSignature.m_vWordBlock[q] = value ^ keyframes[0].scfvSignature.m_vWordBlock[q];
				
		}
		if (hasVar){
			for (int q = 0; q < 512; q++) {
				unsigned int value = 0;
				if (fctPresent[q]) {
					unsigned int* uiBuf = (unsigned int*)(buffer + bufIdx);
					value = uiBuf[0];
					bufIdx += 4;
				}

				desc.scfvSignature.m_vWordVarBlock[q] = value ^ keyframes[0].scfvSignature.m_vWordVarBlock[q];
			}
		}

		desc.scfvSignature.setNorm();

		keyframes.push_back(desc);

	}

}

void CompressedSegmentDescriptor::clear() {
	SegmentDescriptor::clear();

	delete [] globalBuf;  
	globalBuf = 0;
	globalBufSz = 0;
	delete[] localBuf;
	localBuf = 0;
	localBufSz = 0;
	delete[] histoBuf;
	histoBuf = 0;
	histoBufSz = 0;

	dtm.clear();
}


mpeg7cdvs::FeatureList CompressedSegmentDescriptor::readFeatureListFromBinaryAbs(unsigned char* buffer, unsigned int bufferSz, bool readRelevance, unsigned int nrElemGr)
{

	mpeg7cdvs::FeatureList featureList;
	featureList.features.resize(dtm.getNDescr());

	int i;
	int nof_elements = 0;


	nof_elements = nrElemGr * 4;
	//qdescr_size = nof_elements;

	if (featureList.features.size() == 0) {
		return featureList;
	}

	int sh = 7;
	unsigned int offs = 0;
	unsigned char uc = buffer[offs];

	
	int q = 0;
	for (std::vector<mpeg7cdvs::Feature>::iterator f = featureList.features.begin(); f < featureList.features.end(); ++f, q++)
	{
		int read_elems = nof_elements;

		// data coding for absolute descriptor values
		int A = 1; // 0
		int B = 0; // -1
		int C = 2; // +1

		/*
		unsigned short refId = -1;

		
		// use diffs if not from medoid frame
		if (q >= dtm.getNAbsDescr()) {

			// encode index of referenced descriptor (16bit)
			refId = reader.read(16);

		}
		*/

		i = 0;

		while (read_elems)
		{
			if (sh < 0) {
				offs++;
				sh = 7;
				uc = buffer[offs];
			}

			int v = uc & (1 << sh);
			sh--;

			if (v == 0)
			{

				f->qdescr[i] = A;
			}
			else
			{
				if (sh < 0) {
					offs++;
					sh = 7;
					uc = buffer[offs];
				}

				v = uc & (1 << sh);
				sh--;

				if (v == 0)
				{

					f->qdescr[i] = B;
				}
				else
				{

					f->qdescr[i] = C;
				}
			}
			i++;
			read_elems--;
		}

	}



	if (readRelevance)
	{
		for (std::vector<mpeg7cdvs::Feature>::iterator f = featureList.features.begin(); f<featureList.features.end(); ++f)
		{
			if (sh < 0) {
				offs++;
				sh = 7;
				uc = buffer[offs];
			}

			int v = (uc & (1 << sh)) >> sh;

			f->relevance = v;

			sh--;
		}
	}

	featureList.qdescr_size = nof_elements;

	/*
	mpeg7cdva::CompressedFeatureList cfl(featureList, false);

	cfl.relevances.resize(featureList.nFeatures());

	// assign rel info
	int nrel = 0;
	for (int i = 0; i < featureList.nFeatures(); i++) {
		cfl.relevances[i] = featureList.features[i].relevance;
		if (featureList.features[i].relevance>0) nrel++;
	}
	*/

	return featureList;
}



///////////////////////////////// class CdvaImpl //////////////////////////////////


CdvaImpl::CdvaImpl():verboseMode(false),current_op(UNKNOWN),skip_before(4), skip_after(4), drop_frame_th(0.7),cdvsMode(0), shot_ver_th(18),
		shot_cut_th(2.20), max_retrieved(50), calc_desc_sizes(false)
{
	cdvsconfig = NULL;
	cdvsclient = NULL;
	cdvsserver = NULL;
}

CdvaImpl::~CdvaImpl()
{
	delete cdvsclient;
	delete cdvsserver;
	delete cdvsconfig;
}

void CdvaImpl::init(OPERATION op, bool verbose, size_t n_videos, int querybitrate, int refbitrate, bool calcdescsizes, bool rwCompressed, bool optMatch, double set_drop_th, double set_encode_th)
{
	current_op = op;
	verboseMode = verbose;
	calc_desc_sizes = calcdescsizes;

	readWriteCompressed = rwCompressed;

	this->optMatch = optMatch;

	cdvsMode = 0;

	cdvsconfig = CdvsConfiguration::cdvsConfigurationFactory();

	if (current_op == EXTRACT)
		cdvsclient = CdvsClient::cdvsClientFactory(cdvsconfig, cdvsMode); // use ALP

	cdvsserver = CdvsServer::cdvsServerFactory(cdvsconfig);		// used in extract, match and retrieve

	shot_cut_th = 1.98;
	shot_ver_th = 18;
	max_retrieved = 50;
    encode_th = 110;

	optMatch_tau = 0.0;
	optMatch_b = 2;
	optMatch_mode = 1;
	
	switch (querybitrate) {
	case  0:
	case 16:
		skip_before = 4;
		skip_after = 4;
		drop_frame_th = 0.6;
		forceSampleMs = 5000;  //deactivated
		minShotLen = 0;
		minLocalDiff = 1;
		break;
	case 64:
		skip_before = 3;
		skip_after = 3;
		drop_frame_th = 0.5;
		forceSampleMs = 5000; //deactivated
		minShotLen = 0;
		minLocalDiff = 1;
		break;
	case 256:
		skip_before = 2;
		skip_after = 2;
		drop_frame_th = 0.4;
		forceSampleMs = 5000; //deactivated
		minShotLen = 0;
		minLocalDiff = 1;
		break;
	}

	// apply drop threshold if set
	if (set_drop_th > 0) {
		drop_frame_th = set_drop_th;
	}
	// apply encoding threshold if set
	if (set_encode_th > 0) {
		encode_th = set_encode_th;
	}

	// prepare DB (in case retrieval is performed)
	if (current_op == RETRIEVE)
		cdvsserver->createDB(cdvsMode, n_videos);


}

void CdvaImpl::close()
{
	if (current_op == RETRIEVE)
		cdvsserver->clearDB();
}

void CdvaImpl::commitDB()
{
	cdvsserver->commitDB();
	cout << "Index done: stored " << cdvsserver->sizeofDB() << " keyframes."<< endl;
}

/*
* Check valid bitrates
*/
bool CdvaImpl::checkBitrate(int bitrate)
{
	switch (bitrate)
	{
	case   0: return true;
	case  16: return true;
	case  64: return true;
	case 256: return true;
	}
	return false;
}

inline int skipFrames(int framenum, VideoCapture & vc)
{
	int skipped = 0;
	while ((skipped < framenum) && vc.grab())
		skipped++;

	return skipped;
}

inline double getCurrentTimeMillis(VideoCapture & vc)
{
	return ((int) vc.get(CV_CAP_PROP_FOURCC) == FOURCC_MP43)? 100 * vc.get(CV_CAP_PROP_POS_MSEC) : vc.get(CV_CAP_PROP_POS_MSEC);	// fix codec-specific issues
}

inline double getCurrentTimeSeconds(VideoCapture & vc)
{
	return ((int) vc.get(CV_CAP_PROP_FOURCC) == FOURCC_MP43)? 0.1 * vc.get(CV_CAP_PROP_POS_MSEC) : 0.001 * vc.get(CV_CAP_PROP_POS_MSEC);	// fix codec-specific issues
}


/*
* CDVA descriptor extension; may be changed (if needed) to identify a specific implementation.
*/
const char * CdvaImpl::getDescriptorExt(int bitrate)
{
	switch (bitrate)
	{
	case   0: return ".DB.cdva";
	case  16: return ".16K.cdva";
	case  64: return ".64K.cdva";
	case 256: return ".256K.cdva";
	}
	return NULL;
}

//
// return the file extension (in lowercase format)
//
string CdvaImpl::getExt(const string & imageName)
{
	size_t found = imageName.find_last_of('.');
	string s = imageName.substr(found + 1, imageName.length());
	transform(s.begin(),s.end(),s.begin(),::tolower);
	return s;
}


// update the bit counters
void CdvaImpl::updateCounters(const CdvsDescriptor & desc, int total_bit_count, ExtractData & outdata) const
{
	if (calc_desc_sizes)
	{
		static const int hc = 64;	// header bit count - the CDVS header size is 3 + 8 + 1 + 1 +1 +2 (align) + 16 + 16 + 16 = 16x4 =  64 bits (fixed)

		int gc = desc.scfvSignature.compressedNumBits();
		int lc = 6;
		for(std::vector<Feature>::const_iterator f=desc.featurelist.features.begin(); f<desc.featurelist.features.end(); ++f)
		{
			for(unsigned int i=0; i< desc.featurelist.qdescr_size; i++)
			{
					if(f->qdescr[i] == 1)
					{
						lc += 1;
					}
					else
					{
						lc += 2;
					}
			}
		}

		if (cdvsconfig->getParameters(desc.getModeID()).numRelevantPoints > 0)		// relevance bits are present --> update lc
		{
			lc += desc.featurelist.nFeatures();
		}

		int cc = (total_bit_count - (hc + gc + lc));		// the remaining bits belong to coordinate coding

		// cout << hc << " " << gc << " " << lc << " " << cc << " " << endl;

		// update counters

		outdata.header_bit_count += hc;
		outdata.global_bit_count += gc;
		outdata.local_bit_count += lc;
		outdata.coordinate_bit_count += cc;
	}

	++ outdata.n_keyframes;				// update keyframe counter
}



/*
 * Check if this is a key frame (local function).
 */
static void computeColorHistogram(const Mat & src, Mat & r_hist, Mat & g_hist, Mat & b_hist)
{
	if( src.empty() )
		  return;

	/// Separate the image in 3 places ( B, G and R )
	vector<Mat> bgr_planes;
	split( src, bgr_planes );

	/// Establish the number of bins
	int histSize = 32;

	/// Set the ranges ( for B,G,R) )
	float range[] = { 0, 256 } ;
	const float* histRange = { range };

	bool uniform = true; bool accumulate = false;

	/// Compute the histograms:
	calcHist( &bgr_planes[0], 1, 0, Mat(), b_hist, 1, &histSize, &histRange, uniform, accumulate );
	calcHist( &bgr_planes[1], 1, 0, Mat(), g_hist, 1, &histSize, &histRange, uniform, accumulate );
	calcHist( &bgr_planes[2], 1, 0, Mat(), r_hist, 1, &histSize, &histRange, uniform, accumulate );

	/// Normalize the result to [ 0, histImage.rows ]
	normalize(b_hist, b_hist);
	normalize(g_hist, g_hist);
	normalize(r_hist, r_hist);
}

static double diffColorHistogram(const Mat & r, const Mat & g, const Mat & b, const Mat & r_other, const Mat & g_other, const Mat & b_other)
{
	if (r_other.empty() || g_other.empty() || b_other.empty())
	{
		return 1000.0;
	}
	else
	{
		/// compute distance from previous keyframe

		double dist1 = norm(r, r_other, NORM_L2);
		double dist2 = norm(g, g_other, NORM_L2);
		double dist3 = norm(b, b_other, NORM_L2);

		return (dist1 + dist2 + dist3);
	}
}




/*
* CDVA extract implementation
*/
void CdvaImpl::extract(const std::string & descrname, const std::string & videopathname, int mode, ExtractData & outdata) const
{
	// check the filename extension to discriminate images from videos

	string ext = getExt(videopathname);
	bool isVideo = (ext.compare("mp4") == 0||ext.compare("wmv") == 0||ext.compare("flv") == 0||ext.compare("ogv") == 0||ext.compare("mpg") == 0||ext.compare("3gp") == 0
				  ||ext.compare("mov") == 0||ext.compare("mkv") == 0||ext.compare("avi") == 0||ext.compare("m4v") == 0||ext.compare("mpeg") == 0||ext.compare("ogg") == 0);
	bool isImage = (ext.compare("jpg") == 0||ext.compare("jpeg") == 0||ext.compare("bmp") == 0||ext.compare("png") == 0||ext.compare("tiff") == 0||ext.compare("tif") == 0);

	if (!isImage && !isVideo)
	{
		throw CdvaException(string("Unknown file extension: ").append(ext));
	}

	std::ofstream fout;
	fout.open(descrname.c_str(), std::ofstream::binary);
	if (! fout.is_open())
	{
		throw CdvaException(string("Cannot write: ").append(descrname));
	}

	if (isImage)
	{
		Mat greyimage = imread(videopathname, CV_LOAD_IMAGE_GRAYSCALE);
		SegmentDescriptor outputSd;
		CompressedSegmentDescriptor outputComp(cdvsconfig->getParameters(0));
		SegmentDescriptor& output = (readWriteCompressed) ? outputComp : outputSd;

		output.keyframes.resize(1);
		unsigned int descsize = cdvsclient->encode(output.keyframes[0], greyimage.cols, greyimage.rows, greyimage.data);
		output.setStartTimeMs(0);
		output.setEndTimeMs(1000);
		output.setParity(1);

		CdvsDescriptor medoid;
		encodeShot(output, 1000, medoid, fout, outdata);

		outdata.setNumFrames(1);					// 1 picture
		outdata.setNumShots(1);						// 1 shot
		outdata.setVideoDuration(1.0);				// 1 second (conventional value for a picture)

		updateCounters(output.keyframes[0], 8*descsize, outdata);	// update all counters (including the keyframe counter)

		if (verboseMode)
		{
			cout << videopathname << " compressed into 1 keyframe, 1 shot" << endl;
		}

	}
	else if (isVideo)
	{
		VideoCapture videocpp(videopathname);

		if (!videocpp.isOpened())
		{
			throw CdvaException(string("File not found: ").append(videopathname));
		}

		double tot_frames = 0;		// do not trust the video header information: compute the number of frames that can be read

		// MAIN LOOP: encode all frames using CDVA

		Mat colorimage;				// input image
		Mat greyimage;				// converted greyscale image

		Mat r_hist, g_hist, b_hist;	// current histograms (RGB)
		Mat r_last, g_last, b_last;	// previous histograms (RGB)
		Mat r_shot, g_shot, b_shot;	// shot histograms (RGB)

		int c1 = 0;
		int c2 = 0;
		int current_frame = 0;
		unsigned int shotCounter = 1;	// shot counter
		unsigned long currentTimeMs = 0;	// shot start time in milliseconds
		bool isFirst = true;

		SegmentDescriptor outputSd; // create the shot container
		CompressedSegmentDescriptor outputComp(cdvsconfig->getParameters(0));		
		SegmentDescriptor& shot = (readWriteCompressed) ? outputComp : outputSd;		

		shot.setStartTimeMs(currentTimeMs);		// initialize first shot
		shot.setParity(shotCounter);			// initialize first shot
		shot.setEndTimeMs(currentTimeMs);		// will be modified later

		CdvsDescriptor prevDesc;



		do
		{
			bool showImage = false;

			currentTimeMs = ((unsigned long) getCurrentTimeMillis(videocpp));

			tot_frames += (c1 = skipFrames(skip_before, videocpp));		// skip n frames
			if (!videocpp.read(colorimage)) break;		// read one frame
			current_frame = ++tot_frames; 			// must be done before setting c2!
			tot_frames += (c2 = skipFrames(skip_after, videocpp));		// skip n frames

			computeColorHistogram(colorimage, r_hist, g_hist, b_hist);

			if (isFirst)	// encode fist frame
			{
				r_hist.copyTo(r_last);	// last keyframe == this frame
				g_hist.copyTo(g_last);
				b_hist.copyTo(b_last);

				r_hist.copyTo(r_shot);	// last frame in the shot == this frame
				g_hist.copyTo(g_shot);
				b_hist.copyTo(b_shot);

				CdvsDescriptor desc;
				cvtColor(colorimage, greyimage, cv::COLOR_BGR2GRAY);		// convert to luminance only
				int descsize = cdvsclient->encode(desc, greyimage.cols, greyimage.rows, greyimage.data);
				updateCounters(desc, 8*descsize, outdata);	// update all counters (including the keyframe counter)

				prevDesc = desc;

				if (desc.featurelist.nFeatures()>0) shot.keyframes.push_back(desc);			// store the encoded descriptor
				isFirst = false;
				showImage = true;
			}
			// end encoding fist frame
			else	// encode all other frames
			{

				// check if this is a new keyframe
				double distance = diffColorHistogram(r_hist, g_hist, b_hist, r_last, g_last, b_last);
				if (distance > drop_frame_th)
				{
					// check if the new keyframe belongs to the same shot
					double distshot = diffColorHistogram(r_hist, g_hist, b_hist, r_shot, g_shot, b_shot);

					CdvsDescriptor desc;
					cvtColor(colorimage, greyimage, cv::COLOR_BGR2GRAY);		// convert to luminance only
					int descsize = cdvsclient->encode(desc, greyimage.cols, greyimage.rows, greyimage.data);
					updateCounters(desc, 8 * descsize, outdata);	// update all counters (including the keyframe counter)

					
					if ((distshot> shot_cut_th) && (shot.keyframes.size()>0))
					{
						// verify segment boundary using global descriptor
						PointPairs pp = cdvsserver->match(desc, prevDesc, NULL, NULL, MATCH_TYPE_GLOBAL);
						if (pp.global_score <= shot_ver_th) {
							

							r_hist.copyTo(r_shot);
							g_hist.copyTo(g_shot);
							b_hist.copyTo(b_shot);

							// save previous shot container
							shot.setEndTimeMs(currentTimeMs);		// this shot ends where the next starts
							
							// ignore short shots
							if (shot.getEndTimeMs() - shot.getStartTimeMs() >= minShotLen) {

								// process shot
								CdvsDescriptor medoid;


								encodeShot(shot, currentTimeMs, medoid, fout, outdata);

							}


							if (verboseMode)
								cout << "shot: " << shotCounter << ", keyframes: " << shot.keyframes.size() << ", start: " << shot.getStartTimeMs() / 1000.0 << ", stop: " << shot.getEndTimeMs() / 1000.0 << endl;



							// prepare new shot container
							shot.clear();
							shot.setStartTimeMs(currentTimeMs);
							shot.setParity(++shotCounter);

							showImage = true;
						}
					}

					r_hist.copyTo(r_last);
					g_hist.copyTo(g_last);
					b_hist.copyTo(b_last);


					if (desc.featurelist.nFeatures()>0) shot.keyframes.push_back(desc);			// store the encoded descriptor


				
					prevDesc = desc;
				}
			}	// end encoding all other frames

			// displayImageAndHistogram(colorimage, showImage, 32, r_hist, g_hist, b_hist);

		}	while ((c1 == skip_before) && (c2 == skip_after));		// stop if video does not contain other frames

		// save last shot container
		shot.setEndTimeMs((unsigned long) getCurrentTimeMillis(videocpp));
		CdvsDescriptor medoid;

		if (shot.keyframes.size() > 0) {


			encodeShot(shot, shot.getEndTimeMs(), medoid, fout, outdata);

		}

		if (verboseMode)
			cout << "shot: "<< shotCounter << ", keyframes: " << shot.keyframes.size() << ", start: " << shot.getStartTimeMs()/1000.0 << ", stop: " << shot.getEndTimeMs()/1000.0 << endl;

		double time_s = getCurrentTimeSeconds(videocpp);	// normalized time in seconds
		outdata.setVideoDuration(time_s);
		outdata.setNumFrames(tot_frames);
		outdata.setNumShots(shotCounter);
	}	// end else

	outdata.setDescriptorLength((double) fout.tellp());		// real descriptor size (in bytes)
	fout.close();
}

/*
* CDVA match implementation:
*/
double CdvaImpl::match(MatchData & matchResults, const std::string & qdescrname, const std::string & rdescrname, int qmode, int rmode)
{
	if (verboseMode) {
		cout << "matching " << qdescrname << " against " << rdescrname << endl;
	}

	// pairwise matching of frame descriptors
	ShotDescriptorList qDescList, rDescList;

	qDescList.reserve(100);
	rDescList.reserve(100);

	parse(qdescrname, qDescList);
	parse(rdescrname, rDescList);

	if (optMatch) {
		if (optMatch_mode == 1) return match_med1(matchResults, qDescList, rDescList);
		if (optMatch_mode == 2) return match_med2(matchResults, qDescList, rDescList);
		return 0;
	}
	else return match(matchResults, qDescList, rDescList);

}


double CdvaImpl::match(MatchData & matchResults, const ShotDescriptorList& qDescList, const ShotDescriptorList& rDescList)
{
	double maxScore = 0;

	// loop on all shots

	for (int q = 0; q < (int) qDescList.size(); ++q) {
		for (int r = 0; r < (int) rDescList.size(); ++r) {
			for (int i = 0; i < qDescList[q].keyframes.size(); i ++) {
				for (int j = 0; j < rDescList[r].keyframes.size(); j ++) {

					PointPairs ppglobal = cdvsserver->match(qDescList[q].keyframes[i], rDescList[r].keyframes[j], NULL, NULL, MATCH_TYPE_GLOBAL);	// match using Global Descriptor

					if (ppglobal.global_score > ppglobal.global_threshold)
					{
						PointPairs pplocal = cdvsserver->match(qDescList[q].keyframes[i], rDescList[r].keyframes[j], NULL, NULL, MATCH_TYPE_LOCAL); // match using Local Descriptors
						if (pplocal.local_score > pplocal.local_threshold)
						{
							// there is a match
							double score = ppglobal.global_score * pplocal.local_score;

							if (score > maxScore)
								maxScore = score;

							matchResults.setMatchingTime(0.001 * qDescList[q].getStartTimeMs());	// set the query temporal localization (start time)
							matchResults.setMatchingTime(0.001 * qDescList[q].getEndTimeMs());		// set the query temporal localization (end time)
						}
					}
				}
			}
		}
	}

	matchResults.setMatchingScore(maxScore);

	return maxScore;
}


double CdvaImpl::match_med1(MatchData & matchResults, const ShotDescriptorList& qDescList, const ShotDescriptorList& rDescList)
{
	double maxScore = 0;


	// loop on all shots

	for (int q = 0; q < (int)qDescList.size(); ++q) {
		for (int r = 0; r < (int)rDescList.size(); ++r) {

			// match medoids
			PointPairs ppglobal = cdvsserver->match(qDescList[q].keyframes[0], rDescList[r].keyframes[0], NULL, NULL, MATCH_TYPE_GLOBAL);



			if (ppglobal.global_score >= ppglobal.global_threshold*optMatch_tau)
			{


				for (int i = 0; i < qDescList[q].keyframes.size(); i++) {
					for (int j = 0; j < rDescList[r].keyframes.size(); j++) {

						PointPairs ppglobal = cdvsserver->match(qDescList[q].keyframes[i], rDescList[r].keyframes[j], NULL, NULL, MATCH_TYPE_GLOBAL);	// match using Global Descriptor


						if (ppglobal.global_score > ppglobal.global_threshold)
						{
							PointPairs pplocal = cdvsserver->match(qDescList[q].keyframes[i], rDescList[r].keyframes[j], NULL, NULL, MATCH_TYPE_LOCAL); // match using Local Descriptors
							
							if (pplocal.local_score > pplocal.local_threshold)
							{
								// there is a match
								double score = ppglobal.global_score * pplocal.local_score;

								if (score > maxScore)
									maxScore = score;

								matchResults.setMatchingTime(0.001 * qDescList[q].getStartTimeMs());	// set the query temporal localization (start time)
								matchResults.setMatchingTime(0.001 * qDescList[q].getEndTimeMs());		// set the query temporal localization (end time)
							}
						}
					}
				}
			}
		}
	}

	matchResults.setMatchingScore(maxScore);




	return maxScore;
}

double CdvaImpl::match_med2(MatchData & matchResults, const ShotDescriptorList& qDescList, const ShotDescriptorList& rDescList)
{
	double maxScore = 0;
	double minGlobalScoreFactor = optMatch_tau;


	// loop on all shots

	for (int q = 0; q < (int)qDescList.size(); ++q) {
		for (int r = 0; r < (int)rDescList.size(); ++r) {


			int minSegLen = min(qDescList[q].keyframes.size(), rDescList[r].keyframes.size());

			int i = 0;
			int j = 0;

			int matchBudget = 1 + floor(minSegLen / optMatch_b); 
		
			int nMatched = 0;
			double maxGlobalScore = 0;

			bool newI = true;
			if (i >= qDescList[q].keyframes.size()) newI = false;

			while ((matchBudget > 0) && ((i < qDescList[q].keyframes.size()) || (j<rDescList[r].keyframes.size()))) {

				if (newI) {
					for (int jj = j; jj>=0; jj--) {
						PointPairs ppglobal = cdvsserver->match(qDescList[q].keyframes[i], rDescList[r].keyframes[jj], NULL, NULL, MATCH_TYPE_GLOBAL);


						if (ppglobal.global_score >= ppglobal.global_threshold*minGlobalScoreFactor) {
							nMatched++;
							if (ppglobal.global_score > maxGlobalScore) {
								maxGlobalScore = ppglobal.global_score;
							}
						}
					}

				}
				else {
					for (int ii = i; ii>=0; ii--) {
						PointPairs ppglobal = cdvsserver->match(qDescList[q].keyframes[ii], rDescList[r].keyframes[j], NULL, NULL, MATCH_TYPE_GLOBAL);


						if (ppglobal.global_score >= ppglobal.global_threshold*minGlobalScoreFactor) {
							nMatched++;
							if (ppglobal.global_score>maxGlobalScore) {
								maxGlobalScore = ppglobal.global_score;
							}
						}
					}
				}

				matchBudget--;


				if ((j > i) && (j + 1 < rDescList[r].keyframes.size())) {
					if (i + 1 < qDescList[q].keyframes.size()) {
						i++;
						newI = true;
					}
					else {
						j++;
						newI = false;
					}
				}
				else {
					if (j + 1 < rDescList[r].keyframes.size()) {
						j++;
						newI = false;
					}
					else {
						if (i + 1 < qDescList[q].keyframes.size()) {
							i++;
							newI = true;
						}
						else break;
					}
				}
			}

			if (nMatched > 0) {

				for (int i = 0; i < qDescList[q].keyframes.size(); i++) {
					for (int j = 0; j < rDescList[r].keyframes.size(); j++) {

						PointPairs ppglobal = cdvsserver->match(qDescList[q].keyframes[i], rDescList[r].keyframes[j], NULL, NULL, MATCH_TYPE_GLOBAL);	// match using Global Descriptor


						if (ppglobal.global_score > ppglobal.global_threshold)
						{

							PointPairs pplocal = cdvsserver->match(qDescList[q].keyframes[i], rDescList[r].keyframes[j], NULL, NULL, MATCH_TYPE_LOCAL); // match using Local Descriptors

							if (pplocal.local_score > pplocal.local_threshold)
							{

								// there is a match
								double score = ppglobal.global_score * pplocal.local_score;

								if (score > maxScore)
									maxScore = score;

								matchResults.setMatchingTime(0.001 * qDescList[q].getStartTimeMs());	// set the query temporal localization (start time)
								matchResults.setMatchingTime(0.001 * qDescList[q].getEndTimeMs());		// set the query temporal localization (end time)
							}
						}
					}
				}
			}
		}
	}



	matchResults.setMatchingScore(maxScore);

	return maxScore;
}

/*
* CDVA makeindex implementation: can be changed freely;
* you can signal errors at any point of your code using "throw CdvaException(string);"
* the following default implementation of makeindex just appends id + descriptor data to the index file.
*/
void  CdvaImpl::makeindex(const std::string & cdva_descriptor, const std::string & relativepathname)
{
		
	if (!readWriteCompressed) {
		// read descriptor and append it to the DB
		ifstream reader(cdva_descriptor.c_str(), std::ifstream::binary);

		if (!reader.good())
			throw CdvaException(string("Cannot read descriptor: ").append(cdva_descriptor));

		while (SegmentDescriptor::addSegmentToDB(reader, cdvsserver, relativepathname,optMatch) > 0)
		{
		}

	}
	else {

		std::string descriptorId = relativepathname;
		ShotDescriptorList shots;
		parse(cdva_descriptor, shots); // read descriptors from file

		for (int i = 0; i < shots.size(); i++) {
			for (int k = 0; k < shots[i].keyframes.size(); k++) {

				if (optMatch) {

					// stop after part of KFs
					if (k>1 + shots[i].keyframes.size() / optMatch_b) continue;
				
				}

				// ignore descriptors with no local features
				if (shots[i].keyframes[k].featurelist.nFeatures() > 0) {
					cdvsserver->addDescriptorToDB(shots[i].keyframes[k], descriptorId.c_str());	// add image to DB
				}
			}
		}

	}
}

//
// order MatchData elements by descending score
//
bool CdvaImpl::byDescendingScore(const MatchData & m1, const MatchData & m2) {
	return (m1.getScore() > m2.getScore());
}


/*
* CDVA retrieve implementation: can be changed freely;
* you can signal errors at any point of your code using "throw CdvaException(string);"
*/
void CdvaImpl::retrieve(vector<MatchData> & retrievalResults, const std::string & qdescrname, int qmode)
{
	ShotDescriptorList qDescList;

	parse(qdescrname, qDescList);

	retrievalResults.clear();

	map<string, MatchData> matches; // matches per video


	for (int i = 0; i < qDescList.size(); i++) {
		
		for (int j = 0; j < qDescList[i].keyframes.size(); j++) {
					

			vector<RetrievalData> cdvsResults;
			int n = cdvsserver->retrieve(cdvsResults, qDescList[i].keyframes[j], 100);


			// process results (this version saves the top match even if fScore == 0 to avoid empty results)
			for (unsigned int k = 0; k<n; ++k)
			{
				if ((k == 0) || (cdvsResults[k].fScore > 0)) {

					string imgname = cdvsserver->getImageId(cdvsResults[k].index);
					map<string, MatchData>::iterator iter = matches.find(imgname);
					if (iter == matches.end()) {
						MatchData md;
						md.setReferenceID(imgname);
						md.setMatchingScore(cdvsResults[k].fScore);
						md.setMatchingTime(0.001 * qDescList[i].getStartTimeMs());		// (optional) matching time is always related to the query (the reference is known)
						md.setMatchingTime(0.001 * qDescList[i].getEndTimeMs());		// (optional)
						matches[imgname] = md;
					}
					else {
						if (cdvsResults[k].fScore > iter->second.getScore())
						{
							iter->second.setMatchingScore(cdvsResults[k].fScore);
							iter->second.setMatchingTime(0.001 * qDescList[i].getStartTimeMs());	// optional
							iter->second.setMatchingTime(0.001 * qDescList[i].getEndTimeMs());		// optional
						}
					}
				}
			}
		}
	}

	
	// copy remaining match data entries
	for (map<string, MatchData>::iterator it = matches.begin(); it != matches.end(); ++it) {
		retrievalResults.push_back(it->second);
	}

	// check the number of retrieved images - cannot be more than max_retrieved

	if (retrievalResults.size() > max_retrieved)
	{
		// partially sort results by descending score
		partial_sort(retrievalResults.begin(), retrievalResults.begin() + max_retrieved, retrievalResults.end(), byDescendingScore);
		retrievalResults.resize(max_retrieved);		// limit results to the top 50
	}
	else	// size is less or equal than the threshold
	{
		// sort all results by descending score
		sort(retrievalResults.begin(), retrievalResults.end(), byDescendingScore);
	}

	
}


void CdvaImpl::parse(const std::string & descFile, ShotDescriptorList& shotList) {

	ifstream reader(descFile.c_str(), std::ifstream::binary);

	if (!reader.good())
		throw CdvaException(string("Cannot read descriptor: ").append(descFile));

	SegmentDescriptor outputSd;
	CompressedSegmentDescriptor outputComp(cdvsconfig->getParameters(0));
	SegmentDescriptor& shot = (readWriteCompressed) ? outputComp : outputSd;

	while (shot.read(reader, cdvsserver) > 0) {
		shotList.push_back(shot);
		shot.clear();
	}

	reader.close();

}



int CdvaImpl::encodeShot(SegmentDescriptor& shot, unsigned long endTime, CdvsDescriptor& medoid, std::ofstream & fout, ExtractData & outdata) const {

	if (verboseMode) {
		cout << "shot " << shot.getStartTimeMs() << " - " << endTime << " (" << shot.keyframes.size() << " frames) ";
	}

	if (shot.keyframes.size() == 0) {
		if (verboseMode) {
			cout << "no descriptors in shot - ignoring " << endl;
			return 0;
		}
	}

	double** sim = new double*[shot.keyframes.size()];
	for (int i = 0; i < shot.keyframes.size(); ++i) sim[i] = new double[shot.keyframes.size()];

	double maxSim = 0;
	int maxCol = 0;


	// determine medoid

		for (int i = 0; i < shot.keyframes.size(); ++i) {
			double colsum = 0;
			for (int j = 0; j < shot.keyframes.size(); ++j) {
				if (i>j) sim[i][j] = sim[j][i];
				else if (i == j) sim[i][j] = 0;
				else {
					PointPairs pp = cdvsserver->match(shot.keyframes[i], shot.keyframes[j], NULL, NULL, MATCH_TYPE_GLOBAL);
					sim[i][j] = pp.global_score;
				}
				colsum += sim[i][j];
			}
			if (colsum > maxSim) {
				maxSim = colsum;
				maxCol = i;
			}
		}



	std::vector<int> framesToCode;

	// add medoid
	framesToCode.push_back(maxCol);
	medoid = shot.keyframes[maxCol];

	double minSim = maxSim;
	int minIdx = 0;

	do {
		minSim = maxSim;
		minIdx = 0;

		int lastCol = framesToCode[framesToCode.size() - 1];

		bool frameLeft = false;

		for (int j = 0; j < shot.keyframes.size(); ++j) {
			// exclude those already considered
			bool found = false;
			for (int k = 0; k < framesToCode.size(); ++k) {
				if (j == framesToCode[k]) found = true;
			}
			if (!found) {
				// find min of max similarities
				double localMax = -1;
				for (int k = 0; k < framesToCode.size(); k++) {
					if (framesToCode[k] != j) {
						if (sim[framesToCode[k]][j]>localMax) localMax = sim[framesToCode[k]][j];
					}
				}
				if ((localMax > -1) && (localMax < minSim)) {
					minSim = localMax;
					minIdx = j;
					frameLeft = true;
				}
			}
		}

		if (!frameLeft) minSim = encode_th + 1;

		if (minSim < encode_th) {
			framesToCode.push_back(minIdx);
		}

	} while (minSim < encode_th);

	for (int i = 0; i < shot.keyframes.size(); ++i) delete[] sim[i];
	delete[] sim;


	// remove frames of local nr features would exceed 16K
	int totalNrFeatures = 0;
	for (int i = 0; i < framesToCode.size(); i++) {
		totalNrFeatures += shot.keyframes[framesToCode[i]].featurelist.nFeatures();
	}
	while (totalNrFeatures>16000) {
		totalNrFeatures -= shot.keyframes[framesToCode.back()].featurelist.nFeatures();
		framesToCode.erase(framesToCode.end() - 1);
	}


	if (verboseMode) {
		cout << "; encoded " << framesToCode.size() << " frames " << endl;
	}


	// global descriptor differences
	unsigned char* totalBuf = new unsigned char[framesToCode.size() * 16384];
	unsigned int totalBufSz = 0;

	unsigned int uncompressedBufSz = 0;
	unsigned int compressedBufSz = 0;

	
	unsigned int* globaldstatsBin = new unsigned int[2];
	globaldstatsBin[0] = 0;
	globaldstatsBin[1] = 0;

	// for all frames except medoid
	for (int i = 0; i < framesToCode.size(); ++i) {
		if (framesToCode[i] != maxCol){
			SCFVSignature diffSig(medoid.scfvSignature.hasVar(), medoid.scfvSignature.hasBitSelection());
			unsigned int bufSz = 0;
			getDiffSignature(medoid.scfvSignature, shot.keyframes[framesToCode[i]].scfvSignature, diffSig, bufSz, totalBuf + totalBufSz, globaldstatsBin);
			totalBufSz += bufSz;
			uncompressedBufSz += shot.keyframes[framesToCode[i]].scfvSignature.compressedNumBits() / 8;
		}
	}

	if (totalBufSz>0) {



		// encode diff
		evx::bitstream src(totalBufSz * 8 * 2);
		evx::bitstream dst(totalBufSz * 8 * 2);

		src.write_bytes(totalBuf, totalBufSz);

		evx::entropy_coder coder;
		coder.encode(&src, &dst, true);
		coder.finish_encode(&dst);

		compressedBufSz = dst.query_byte_occupancy();

		if (readWriteCompressed) {
			CompressedSegmentDescriptor& csd = (CompressedSegmentDescriptor&) shot;

			unsigned char* buffer = csd.allocateGlobalBuffer(compressedBufSz);
			csd.setGlobalUncompressedBufSz(totalBufSz);

		
			evx::uint32 actBufSz = (evx::uint32)(compressedBufSz & 0xffffffff);
			dst.read_bytes(buffer, &actBufSz);
		}


	}
	else {
		compressedBufSz = 0;

		if (readWriteCompressed) {
			CompressedSegmentDescriptor& csd = (CompressedSegmentDescriptor&)shot;
			csd.setGlobalUncompressedBufSz(totalBufSz);

		}
	}

	delete[] totalBuf;

	
	unsigned int localBufSz = 0;
	int localCompressedBufSz = 0;
	unsigned int* localdstatsDiff = new unsigned int[3];


	// local descriptor absolute encoding 

	unsigned int localBufSzAbs = 0;
	unsigned char* localBuf = new unsigned char[shot.keyframes.size() * 400 * 1000]; 

	unsigned int* localdstatsAbs = new unsigned int[3];
	localdstatsAbs[0] = 0;
	localdstatsAbs[1] = 0;
	localdstatsAbs[2] = 0;

	localDescCodingAbs(shot, framesToCode, localBufSzAbs, localBuf, maxCol, localdstatsAbs);

	int localCompressedBufSzAbs = 0;

	if (localBufSzAbs>0) {
		// encode diff
		evx::bitstream src(localBufSzAbs * 8);
		evx::bitstream dst(localBufSzAbs * 8 * 2);

		src.write_bytes(localBuf, localBufSzAbs);

		evx::entropy_coder coder;
		coder.encode(&src, &dst, true);
		coder.finish_encode(&dst);

		localCompressedBufSzAbs = dst.query_byte_occupancy();

		if (readWriteCompressed) {
			CompressedSegmentDescriptor& csd = (CompressedSegmentDescriptor&)shot;
			csd.setLocalUncompressedBufSz(localBufSzAbs);

			unsigned char* buffer = csd.allocateLocalBuffer(localCompressedBufSzAbs);

			evx::uint32 actBufSz = (evx::uint32)(localCompressedBufSzAbs & 0xffffffff);
			dst.read_bytes(buffer, &actBufSz);
		}
	}
	else {
		localCompressedBufSzAbs = 0;
	}
	
	delete[] localBuf;

	

	// ---

	unsigned int localUncompressed = 0;

	for (int i = 0; i < framesToCode.size(); ++i) {
		localUncompressed += 6;
		for (std::vector<Feature>::const_iterator f = shot.keyframes[framesToCode[i]].featurelist.features.begin(); f<shot.keyframes[framesToCode[i]].featurelist.features.end(); ++f)
		{
			for (unsigned int j = 0; j< shot.keyframes[framesToCode[i]].featurelist.qdescr_size; j++)
			{
				if (f->qdescr[j] == 1)
				{
					localUncompressed += 1;
				}
				else
				{
					localUncompressed += 2;
				}
			}
		}
		if (cdvsconfig->getParameters(shot.keyframes[framesToCode[i]].getModeID()).numRelevantPoints > 0)		// relevance bits are present --> update lc
		{
			localUncompressed += shot.keyframes[framesToCode[i]].featurelist.nFeatures();
		}	
	}

	

	delete[] localdstatsDiff;
	delete[] localdstatsAbs;
	delete[] globaldstatsBin;



	if (!readWriteCompressed) {
		SegmentDescriptor filteredShot;

		filteredShot.setStartTimeMs(shot.getStartTimeMs());
		filteredShot.setEndTimeMs(shot.getEndTimeMs());
		filteredShot.setParity(shot.getParity());

		for (int i = 0; i < framesToCode.size(); ++i) {

			int pos = fout.tellp();


			filteredShot.keyframes.push_back(shot.keyframes[framesToCode[i]]);


			if (calc_desc_sizes) updateCounters(shot.keyframes[framesToCode[i]], 8 * ((int)fout.tellp() - pos), outdata);
		}

		filteredShot.write(fout);
	}
	else {
		std::vector <mpeg7cdvs::CdvsDescriptor> allKeyframes = shot.keyframes;
		shot.keyframes.clear();
		for (int i = 0; i < framesToCode.size(); ++i) {
			shot.keyframes.push_back(allKeyframes[framesToCode[i]]);
		}

		shot.write(fout);

	
	}


	return framesToCode.size();

}

void CdvaImpl::getDiffSignature(SCFVSignature& medoid, SCFVSignature& other, SCFVSignature& diffSig, unsigned int& bufferSize, unsigned char* buffer, unsigned int* globaldstatsBin) const {

	int nrDescM = 0, nrDescO = 0;
	int offs = 8; // the interesting block starts at byte 8

	// determine differences that need to be represented
	int nrDiff = 0;
	for (int q = 0; q < 512; q++) {
		if ((medoid.m_vWordBlock[q]>0) || (other.m_vWordBlock[q]>0)) {

			unsigned int diff = 0;
			unsigned int vardiff = 0;
			int nBitsDiff = 0;

			int present = 1;

			//if (setC + setP > 1) {
				diff = (medoid.m_vWordBlock[q]) ^ (other.m_vWordBlock[q]);
				vardiff = (medoid.m_vWordVarBlock[q]) ^ (other.m_vWordVarBlock[q]);
				if ((diff == 0) && (vardiff == 0)) present = 0;
				else nBitsDiff = nBitsSet(diff);
			/*}
			else {
				if (setC > 0) {  
					diff = medoid.m_vWordBlock[q];
					vardiff = medoid.m_vWordVarBlock[q];
				}
				else {
					vardiff = other.m_vWordVarBlock[q];
				}
				nBitsDiff = nBitsSet(diff);
			}*/
			

			diffSig.m_vWordBlock[q] = diff;
			diffSig.m_vWordVarBlock[q] = vardiff;
			if (present) {
				nrDiff++;
			
				// TESTING ONLY
				globaldstatsBin[1] += nBitsDiff;
				globaldstatsBin[0] += 32-nBitsDiff;

			}

		}
	}
	diffSig.setVisited();
	diffSig.setNorm();

	bufferSize = 64 + nrDiff * 3 + nrDiff*(diffSig.hasBitSelection() ? 0 : 1) + nrDiff * 4 * (diffSig.hasVar() ? 1 : 0);
	// write diff to buffer 
	//unsigned char* buffer = new unsigned char[bufferSize];

	// present
	int presIdx = -1;
	for (int q = 0; q < 512; q++) {
		char sh = q % 8;
		if (q % 8 == 0) {
			presIdx++;
			buffer[presIdx] = 0;
		}

		buffer[presIdx] += (((diffSig.m_vWordBlock[q]>0) || (diffSig.m_vWordVarBlock[q]>0)) ? 1 : 0) << sh;


	}

	unsigned int* uiBuf = 0;

	// means
	offs = 64;
	for (int q = 0; q < 512; q++) {

		if ((diffSig.m_vWordBlock[q]>0) || (diffSig.m_vWordVarBlock[q]>0)) {
			uiBuf = (unsigned int*)(buffer + offs);
			uiBuf[0] = diffSig.m_vWordBlock[q];
			offs += 4;
		}

	}


	int meanOff = offs;
	// variances
	if (diffSig.hasVar()) {

		for (int q = 0; q < 512; q++) {

			if ((diffSig.m_vWordBlock[q]>0) || (diffSig.m_vWordVarBlock[q]>0)) {
				uiBuf = (unsigned int*)(buffer + offs);
				uiBuf[0] = diffSig.m_vWordVarBlock[q];
				offs += 4;
			}

		}
	}
	int totalSz = offs;

	assert(totalSz == bufferSize);

}



void CdvaImpl::localDescCodingAbs(SegmentDescriptor& shot, std::vector<int>& framesToCode, unsigned int& bufferSize, unsigned char* buffer, int medoidIdx, unsigned int* localdstatsAbs) const {
	int numConsideredDesc = shot.keyframes.size();

	// init feature list from ref
	std::vector<RefFeature> refFeatureList;

	mpeg7cdvs::CdvsDescriptor& medoidDesc = shot.keyframes[medoidIdx];

	if (medoidDesc.featurelist.nFeatures() == 0) {
		cout << "no features in medoid desc " << endl;
		return;
	}

	for (int q = 0; q < medoidDesc.featurelist.nFeatures(); q++) {
		RefFeature rf(medoidDesc.featurelist.features[q]);
		rf.timesPresent.push_back(0); // medoid frames is frame # 0
		refFeatureList.push_back(rf);
	}

	std::vector<int> ldOffsets; // offsets in local descriptor list for each frame
	std::vector<int> localFrameOrder;

	ldOffsets.resize(framesToCode.size());
	ldOffsets[0] = 0;
	localFrameOrder.push_back(0);//Medoid

	// start around refFrame
	int sign = 1;
	int frameOutIdx = 0; // frame index in the order to be serialised
	for (int q = 1, sign = 1; q < numConsideredDesc; sign *= -1) {

		int qi = (medoidIdx + q*sign);//Medoid

		bool skip = false;

		if (qi < 0) skip = true;
		if (qi > numConsideredDesc - 1) skip = true;


		// get output frame index
		frameOutIdx = 1;
		int j = 1;

		if (framesToCode.size() == 1) {
			skip = true;
			break;
		}

		while (framesToCode[j] != qi) {
			j++;
			frameOutIdx++;
			if (j == framesToCode.size()) {
				skip = true;
				break;
			}
		}

		if (skip) {
			if (sign < 0) q++;
			continue;
		}

		ldOffsets[frameOutIdx] = refFeatureList.size();
		localFrameOrder.push_back(frameOutIdx);

		FeatureList& fl = shot.keyframes.at(qi).featurelist;
		int flsz = fl.nFeatures();

		int kept = 0;
		int refd = 0;

		// determine most similar features in any of the features processed for far 
		for (int rr = 0; rr < flsz; rr++) {
			int minDiff = RefFeature::descrLength * 3;
			int minIdx = -1;
			RefFeature rf(fl.features[rr]);

			rf.timesPresent.push_back(frameOutIdx);
			int* rrA = fl.features[rr].qdescr;

			assert(refFeatureList.size() <= INT_MAX);

			for (int r = refFeatureList.size() - 1; r >= 0; r--) {
				int qdiff = 0;
				int qdiffArray[RefFeature::descrLength];
				if (!refFeatureList[r].refOnly) {// do not match against descriptors that are just refs
					int* rA = refFeatureList[r].qdescr;

					qdiff = CompressedFeatureList::getDistance((const unsigned char*)refFeatureList[r].compressed, (const unsigned char*)rf.compressed, 32);

					// stop if very similar descriptor found
					if (qdiff < this->minLocalDiff) {
						// only set if not more than 16 refs from this frame
						bool doRef = true;
						if (refFeatureList[r].timesPresent.size()>15) {

							int sameFrameCount = 0;
							for (int kk = 0; kk < refFeatureList[r].timesPresent.size(); kk++) {
								if (refFeatureList[r].timesPresent[kk] == frameOutIdx) sameFrameCount++;
							}
							if (sameFrameCount>14) doRef = false;
						}
						if (doRef) {
							minDiff = -1;
							minIdx = r;

							// FOR TESTING: set original descriptor to referenced one
							for (int jj = 0; jj < 128; jj++)	{
								fl.features[rr].descr[jj] = refFeatureList[r].descr[jj];
								fl.features[rr].qdescr[jj] = refFeatureList[r].qdescr[jj];
							}

							break;
						}
						else {
							minDiff = qdiff;
							minIdx = r;

						}
					}
					// optimisation: stop at sufficiently similar one to speed up
					else if (qdiff < this->minLocalDiff * 2) {
						minDiff = qdiff;
						minIdx = r;
						break;
					}
					else {
						if (qdiff < minDiff) {
							minDiff = qdiff;
							minIdx = r;

						}
					}
				}
			}

			// found a very similar descriptor
			if (minDiff < 0) {
				rf.refOnly = true;

				refd++;

				// add time to referenced descriptor 

				refFeatureList[minIdx].timesPresent.push_back(rf.timesPresent.at(0));

			}
			else {
				rf.refOnly = false;
				kept++;
			}

			rf.refFeatureIndex = minIdx;
			rf.diff = minDiff;

			// determine diff vector
			int* rA = refFeatureList[minIdx].qdescr;
			for (int s = 0; s < RefFeature::descrLength; s++) {

				rf.qdescr[s] = rrA[s]; 

			}

			refFeatureList.push_back(rf);

		}


		if (sign < 0) q++;
	}


	int totalNrFeatures = 0;
	for (int q = 0; q < refFeatureList.size(); q++) {
		if (!refFeatureList[q].refOnly) {
			totalNrFeatures++;

		}
	}

	assert(totalNrFeatures <= 16000);


	/*
	int lastFrame = localFrameOrder.size() - 1;

	// remove frames if nr local features exceeded
	while (totalNrFeatures>16000) {
		if (verboseMode) cout << "exceeed max nr features - reducing " << std::endl;

		int idx = localFrameOrder[lastFrame];

		int nRemoved = 0;
		// remove local feature
		for (int q = 0; q < refFeatureList.size(); q++) {
			if (refFeatureList[q].refOnly) continue;

			std::vector<unsigned int>::iterator titer = refFeatureList[q].timesPresent.begin();
			while (titer != refFeatureList[q].timesPresent.end()) {

				if (*titer == idx) {
					titer = refFeatureList[q].timesPresent.erase(titer);
				}
				else titer++;
			}
			// if this was the only frame, remove feature
			if (refFeatureList[q].timesPresent.size() == 0) {
				refFeatureList[q].refOnly = true;
				refFeatureList[q].refFeatureIndex = -1;
				nRemoved++;
			}
		}

		totalNrFeatures -= nRemoved;

		if (verboseMode) cout << "nr features is now " << totalNrFeatures << endl;
		lastFrame--;
		framesToCode.erase(framesToCode.end()-1);

	}
	*/


	// calculate offset for renumbering after skipped descriptors
	int refOffs = 0; // feature references are reindexed (as refOnly features are skipped from the list)
	int nActualFeatures = 0;

	std::vector<int> refOffsList;

	for (int q = 0; q < refFeatureList.size(); q++) {

		if (refFeatureList[q].refOnly) {
			refOffs++;

			refOffsList.push_back(refOffs);
		}
		else {
			nActualFeatures++;

			refOffsList.push_back(refOffs);

			// possibly renumber if not a feature from medoid frame
			if (q >= shot.keyframes[framesToCode[0]].featurelist.nFeatures()) {
				refFeatureList[q].refFeatureIndex -= refOffsList[refFeatureList[q].refFeatureIndex];
			}
		}
	}

	// - descriptor and difference descriptors (to be compressed)

	char data = 0;
	int bitIdx = 7;
	bufferSize = 0;

	int nrElem = cdvsconfig->getParameters(0).numberOfElementGroups*4;

	for (int q = 0; q < refFeatureList.size(); q++) {

		if (!refFeatureList[q].refOnly) {

			int* dataArray = refFeatureList[q].qdescr;

			// data coding for absolute descriptor values
			int A = 1; // 0
			int B = 0; // -1
			int C = 2; // +1

			for (int s = 0; s < nrElem; s++) {


				if (dataArray[s] == A) {
					// A -> 0
					//cout << "0";
					localdstatsAbs[A]++;

					bitIdx--;
					if (bitIdx == -1) {
						buffer[bufferSize] = data;
						bufferSize++;
						bitIdx = 7;
						data = 0;
					}
				}
				else {
					// B -> 10
					// C -> 11
					//cout << "1";

					data += 1 << bitIdx;
					bitIdx--;
					if (bitIdx == -1) {
						buffer[bufferSize] = data;
						bufferSize++;
						bitIdx = 7;
						data = 0;
					}

					if (dataArray[s] == B) {
						//cout << "0";
						localdstatsAbs[B]++;

						bitIdx--;
						if (bitIdx == -1) {
							buffer[bufferSize] = data;
							bufferSize++;
							bitIdx = 7;
							data = 0;
						}
					}
					else {
						//cout << "1";
						localdstatsAbs[C]++;

						data += 1 << bitIdx;
						bitIdx--;
						if (bitIdx == -1) {
							buffer[bufferSize] = data;
							bufferSize++;
							bitIdx = 7;
							data = 0;
						}

					}
				}
			}



		}
	}

	// - relevance
	// only serialise if indicated in header
	if (shot.keyframes[0].getRelevanceBitsPresent()) {
		for (int q = 0; q < refFeatureList.size(); q++) {


			if (refFeatureList[q].relevance>0) data += 1 << bitIdx;

			bitIdx--;
			if (bitIdx == -1) {
				buffer[bufferSize] = data;
				bufferSize++;
				bitIdx = 7;
				data = 0;
			}


		}
	}

	// write rest
	if (bitIdx<7) {

		buffer[bufferSize] = data;
		bufferSize++;
	}

	if (readWriteCompressed) {
		generateDTM((CompressedSegmentDescriptor&)shot, framesToCode, medoidIdx, refFeatureList, ldOffsets, localFrameOrder, nActualFeatures);

		encodeCoordinates((CompressedSegmentDescriptor&)shot, framesToCode, medoidIdx, refFeatureList, ldOffsets);
	}


}


void CdvaImpl::generateDTM(CompressedSegmentDescriptor& shot, std::vector<int>& framesToCode, int medoidIdx, std::vector<RefFeature> refFeatureList, std::vector<int> ldOffsets, std::vector<int> localFrameOrder, int totalNrFeatures) const {



	DescriptorTimeMap& dtm = shot.getDtm();
	
	dtm.clear();

	dtm.init(totalNrFeatures, shot.keyframes[framesToCode[0]].featurelist.nFeatures(), framesToCode.size());

	dtm.ldOffsets = ldOffsets;

	// fill descriptor time map (keep original order)

	
	for (int k = 0; k < framesToCode.size(); k++) {
		
		int i = 0;
		int ja = 0;


		std::vector<int> reindexed;
		bool offsetSet = false;

		for (int j = 0; j < ldOffsets[k] + shot.keyframes[framesToCode[k]].featurelist.nFeatures(); j++) {

			if (!refFeatureList[j].refOnly) {
				reindexed.push_back(ja);
			}
			else {
				reindexed.push_back(-1);
			}

			if (j >= ldOffsets[k]) {

				if (refFeatureList[j].refOnly) {
					if (refFeatureList[j].refFeatureIndex > -1) {
						dtm.add(reindexed[refFeatureList[j].refFeatureIndex], k);
					}
					else {
						cout << "feature removed" << endl;
					}
				}
				else {
					dtm.add(ja, k);
					if (!offsetSet) {
						dtm.ldOffsets[k] = ja;
						offsetSet;
					}

				}
			}

			if (!refFeatureList[j].refOnly) ja++;

		}
	
	}

}

void CdvaImpl::encodeCoordinates(CompressedSegmentDescriptor& shot, std::vector<int>& framesToCode, int medoidIdx, std::vector<RefFeature> refFeatureList, std::vector<int> ldOffsets) const {
	
	unsigned char* buffer = shot.allocateHistoBuffer(10000 * framesToCode.size());
	
	BitOutputStream bos(buffer,shot.getHistoBufSz());

	for (int i = 0; i < framesToCode.size(); i++) {
		CdvsDescriptor& cdvs = shot.keyframes[framesToCode[i]];

		CoordinateCoding cc(cdvsconfig->getParameters(0));

		cc.generateHistogramMap(refFeatureList, cdvs.featurelist, ldOffsets[i], i, shot.getDtm(), cdvs.featurelist.nFeatures(), cdvs.featurelist.imageWidth, cdvs.featurelist.imageHeight);
		cc.toBinary(bos, false);


	}

	shot.setHistoBufSz( bos.produced());

	bos.close();


}

