/*
The copyright in this software is being made available under this MPEG Reference Software Copyright License. This software may be subject to other third party and contributor rights, including patent rights, and no such rights are granted under this license.

Copyright (c) 2016-2017, Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
*	Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
*	Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
*	Neither the names of Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms nor the names of their contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#pragma once

#include <string>
#include <vector>
#include <map>

#include "cdva.h"
#include "CdvaException.h"
#include "LogManager.h"
#include "Buffer.h"
#include "CdvsInterface.h"

#include "SCFVIndex.h"
#include "DescriptorTimeMap.h"
#include "CompressedFeatureList.h"

/**
* @namespace mpeg7cdva
* Namespace used to encapsulate all MPEG-7 CDVA declarations that are visible when
* the CDVA headers (in particular cdva.h) are included.
*/
namespace mpeg7cdva
{

	enum OPERATION { UNKNOWN, EXTRACT, MATCH, RETRIEVE };
	

	/**
	 * @class SegmentDescriptor
	 * A container for CdvsDescriptor instances belonging to the same video segment.
	 */
	class SegmentDescriptor
	{

	protected:
		unsigned long startTime;
		unsigned long endTime;
		unsigned int parity;

		static const int SHOT_HEADER_SIZE = 10;		// header size in bytes

		size_t getSize() const;		// get the size in bytes of encoded keyframes when written into a buffer

	public:
		SegmentDescriptor();
		virtual ~SegmentDescriptor();

		std::vector <mpeg7cdvs::CdvsDescriptor> keyframes;	///< keyframes belonging to this shot

		void setParity(unsigned int value);					///< Set the shot parity
		unsigned int getParity() const;						///< Get the shot parity
		void setStartTimeMs(unsigned long position_ms);		///< Set the start time of the shot in milliseconds
		void setEndTimeMs(unsigned long position_ms);		///< Set the start time of the frame in milliseconds
		unsigned long getStartTimeMs() const;				///< Get the shot start time in milliseconds
		unsigned long getEndTimeMs() const;					///< Get the frame start time in milliseconds
		virtual size_t write(std::ofstream & fout) const;			///< write (appending) this segment to a file
		virtual size_t read(std::ifstream & fin, mpeg7cdvs::CdvsServer * cdvsserver);		///< read (from the current position) this segment from a file
		static size_t addSegmentToDB(std::ifstream & fin, mpeg7cdvs::CdvsServer * cdvsserver, const std::string & relativepathname,bool optMatch); ///< read (from the current position) this segment from a file and store it into the CdvsServer DB
		bool empty() const;				///< returns true if the segment is empty
		virtual void clear();					///< clear the shot container
	};

	/**
	 * @class CompressedSegmentDescriptor
	 * Extension of segment descriptor container to also hold compressed data for a segment
	 */
	class CompressedSegmentDescriptor : public SegmentDescriptor
	{
	protected:
		unsigned char* globalBuf;   ///< buffer for differential global descriptors
		unsigned char* localBuf;    ///< buffer for local descriptors
		unsigned char* histoBuf;    ///< buffer for histogram descriptors
		unsigned long globalBufSz;  ///< size of encoded global buffer
		unsigned long localBufSz;   ///< size of encoded local buffer
		unsigned long histoBufSz;   ///< size of encoded histogram buffer
		unsigned long globalUncompressedBufSz;  ///< size of unencoded global buffer
		unsigned long localUncompressedBufSz;  ///< size of unencoded local buffer

		DescriptorTimeMap dtm;		///< map for frame - descriptor association

		const mpeg7cdvs::Parameters& params; ///< parameters needed to invoke histogram decoding

		static const int SHOT_HEADER_SIZE = 35;		// header size in bytes

		void decode(char* globalMedBuf, unsigned int globalMedBufSz, std::vector<unsigned int>& nLocalDesc,				///< decode compressed descriptor structure
			unsigned char cdvsDescriptorMode, unsigned short origImgX, unsigned short origImgY, unsigned short histoMapSzX, unsigned short histoMapSzY,
			bool hasBitSel, bool hasVar, bool hasRelB);

		void reconstructGlobalDifferences(char* uncompressedData, unsigned int uncompressedSz, int nrGDesc, bool hasVar, bool globalHasBitSelection, std::vector<unsigned int>& nLocalDesc);

		mpeg7cdvs::FeatureList readFeatureListFromBinaryAbs(unsigned char* buffer, unsigned int bufferSz, bool readRelevance, unsigned int nrElemGr);

		inline int numGlobFctPresent(const mpeg7cdvs::CdvsDescriptor& desc) const {

			int numFctPresent = 0;
			int offs = 8; // the interesting block starts at byte 8
			for (int q = 0; q < 64; q++) {
				numFctPresent += nBitsSet(desc.buffer.data()[offs + q]);

			}
			return numFctPresent;
		}

		inline int nBitsSet(unsigned char byte) const {
			return ((byte & 0x01) ? 1 : 0) + ((byte & 0x02) ? 1 : 0) + ((byte & 0x04) ? 1 : 0) + ((byte & 0x08) ? 1 : 0) +
				((byte & 0x10) ? 1 : 0) + ((byte & 0x20) ? 1 : 0) + ((byte & 0x40) ? 1 : 0) + ((byte & 0x80) ? 1 : 0);
		}


	public:
		CompressedSegmentDescriptor(const mpeg7cdvs::Parameters& params);
		virtual ~CompressedSegmentDescriptor();

		virtual size_t write(std::ofstream & fout) const;			///< write (appending) this segment to a file
		virtual size_t read(std::ifstream & fin, mpeg7cdvs::CdvsServer * cdvsserver);		///< read (from the current position) this segment from a file

		virtual unsigned char* allocateLocalBuffer(unsigned long sz); 
		virtual unsigned char* allocateGlobalBuffer(unsigned long sz);
		virtual unsigned char* allocateHistoBuffer(unsigned long sz);

		virtual unsigned char* getLocalBuffer() { return localBuf; }
		virtual unsigned char* getGlobalBuffer() { return globalBuf; }
		virtual unsigned char* getHistoBuffer() { return histoBuf; }

		virtual unsigned long getLocalBufSz() { return localBufSz;  }
		virtual unsigned long getGlobalBufSz() { return globalBufSz; }
		virtual unsigned long getHistoBufSz() { return histoBufSz; }

		virtual void setHistoBufSz(unsigned long sz) { histoBufSz = sz; }

		virtual void setLocalUncompressedBufSz(unsigned long sz) { localUncompressedBufSz = sz;  }
		virtual void setGlobalUncompressedBufSz(unsigned long sz) { globalUncompressedBufSz = sz; }

		virtual DescriptorTimeMap& getDtm() { return dtm; }

		virtual void clear();

	};


	typedef std::vector<SegmentDescriptor> ShotDescriptorList;		///< vector of shots descriptors of a video


	class RefFeature;


	/**
	* @class CdvaImpl
	* A CDVA implementation based on multiple CDVS descriptors
	*/
	class CdvaImpl
	{
	private:
		void updateCounters(const  mpeg7cdvs::CdvsDescriptor & desc, int total_bit_count, ExtractData & outdata) const;
	protected:
		mpeg7cdvs::CdvsConfiguration * cdvsconfig;
		mpeg7cdvs::CdvsClient * cdvsclient;
		mpeg7cdvs::CdvsServer * cdvsserver;

		bool verboseMode;		///< verbose mode indicator
		OPERATION current_op;	///< the current operation
		int cdvsMode;			///< the CDVS mode that will be used to encode keyframe descriptors
		int skip_before;		///< number of video frames to skip before decoding one
		int skip_after;			///< number of video frames to skip after decoding one
		double drop_frame_th;	///< drop frame threshold
		double shot_cut_th;		///< shot cut threshold
		double shot_ver_th;		///< shot verification threshold
		double encode_th;		///< threshold for encoding other frames than median
		size_t max_retrieved;	///< maximum number of retrieved images
		bool calc_desc_sizes;   ///< calculate size of descriptor components
		int forceSampleMs;		///< enforce a sample every x milliseconds
		int minShotLen;			///< minimum shot length
		int minLocalDiff;		///< min local difference for lossy local descriptor coding (if diff is larger, descriptor is encoded, otherwise replaced by reference)
		bool readWriteCompressed;///< read and write compressed descriptors
		bool optMatch;			///< approx. matching of medoid-based descriptors
		int optMatch_mode;		///< mode for optimised matching (1 or 2)
		double optMatch_tau;	///< factor to define threshold to continue matching (opt matching modes 1 and 2)
		double optMatch_b;		///<  divider for fraction of key frames to check to decide for match (opt matching mode 2)

		static bool byDescendingScore(const MatchData & m1, const MatchData & m2);	// order data by descending score

		virtual void parse(const std::string & descFile, ShotDescriptorList& shotList);

		virtual double match(MatchData & matchResults, const ShotDescriptorList& qDescList, const ShotDescriptorList& rDescList);

		/** matching using medoid - variant 1
		  * check medoids against min threshold, then perform full match
		  */
		virtual double match_med1(MatchData & matchResults, const ShotDescriptorList& qDescList, const ShotDescriptorList& rDescList);

		/** matching using medoid - variant 2
		* check medoids against min threshold, then iteratively match frames
		*/
		virtual double match_med2(MatchData & matchResults, const ShotDescriptorList& qDescList, const ShotDescriptorList& rDescList);


		virtual int encodeShot(SegmentDescriptor& shot, unsigned long endTime, mpeg7cdvs::CdvsDescriptor& medoid, std::ofstream & fout, ExtractData & outdata) const;  // returns nr frames encoded

		virtual void getDiffSignature(mpeg7cdvs::SCFVSignature& medoid, mpeg7cdvs::SCFVSignature& other, mpeg7cdvs::SCFVSignature& diffSig, unsigned int& bufferSize, unsigned char* buffer, unsigned int* globaldstatsBin) const;


		/** lossy local descriptor coding, applying ABAC to absolute descriptors
		  * @param localdstats for gathering encoding statistics only
		  */
		virtual void localDescCodingAbs(SegmentDescriptor& shot, std::vector<int>& framesToCode, unsigned int& bufferSize, unsigned char* buffer, int medoidIdx, unsigned int* localdstats) const;

		/** generate frame - descriptor map
		*/
		void generateDTM(CompressedSegmentDescriptor& shot, std::vector<int>& framesToCode, int medoidIdx, std::vector<RefFeature> refFeatureList, std::vector<int> ldOffsets, std::vector<int> localFrameOrder, int totalNrFeatures) const;


		/** encode coordinates of local descriptors of as one block
		  */
		virtual void encodeCoordinates(CompressedSegmentDescriptor& shot, std::vector<int>& framesToCode, int medoidIdx, std::vector<RefFeature> refFeatureList, std::vector<int> ldOffsets) const;


		static inline int nBitsSet(unsigned char byte) {
			return ((byte & 0x01) ? 1 : 0) + ((byte & 0x02) ? 1 : 0) + ((byte & 0x04) ? 1 : 0) + ((byte & 0x08) ? 1 : 0) +
				((byte & 0x10) ? 1 : 0) + ((byte & 0x20) ? 1 : 0) + ((byte & 0x40) ? 1 : 0) + ((byte & 0x80) ? 1 : 0);
		}

		static inline int nBitsSet(unsigned int word) {
			unsigned char* ba = (unsigned char*)&word;
			return nBitsSet(ba[0]) + nBitsSet(ba[1]) + nBitsSet(ba[2]) + nBitsSet(ba[3]);

		}

		/*
		 * Get the extension of a file name.
		 * @param imageName the original image/video name;
		 * @return the extension of the name;
		 */
		static std::string getExt(const std::string & imageName);

	public:
		/**
		 * check if the given bitrate is one of the standard values defined in the CDVA evaluation framework.
		 * @param bitrate the bitrate in Kilo-byte per second (KB/s)
		 * @return true if valid
		 */
		static bool checkBitrate(int bitrate);

		/**
		* get the file extension corresponding to the given bitrate.
		* @param bitrate the bitrate in Kilo-byte per second (KB/s)
		* @return the file extension
		*/
		static const char * getDescriptorExt(int bitrate);

		CdvaImpl();
		virtual ~CdvaImpl();

		/**
		* initialization method - called once before processing videos.
		* @param op one of EXTRACT, MATCH, RETRIEVE
		* @param verbose when set, more information is provided
		* @param n_videos the number of videos to be processed
		* @param querybitrate the query encoding bitrate (one of 0,16,64,256)
		* @param refbitrate the reference encoding bitrate (one of 0,16,64,256)
		* @param calcdescsizes if true, the size of individual elements of the CDVS descriptor is reported in the output log files
		* @param rwCompressed if true, compressed descriptor representations are used for serialising and parsing
		* @param optMatch if true, use approximate matching of medoid-based descriptors
		* @param set_drop_th value to override default drop frame threshold
		* @param set_encode_th value to override default encode frame threshold
		*/
		virtual void init(OPERATION op, bool verbose, size_t n_videos, int querybitrate, int refbitrate = 0, bool calcdescsizes = false, bool rwCompressed = false, bool optMatch = false, double set_drop_th = -1, double set_encode_th = -1);

		/**
		 * Video processing method - called once for each video in the list.
		 * @param descrname output descriptor pathname
		 * @param videopathname input video stream pathname
		 * @param bitrate encoding bitrate (one of 0,16,64,256)
		 * @param outdata the container for output data
		 */
		void extract(const std::string & descrname, const std::string & videopathname, int bitrate, ExtractData & outdata) const;

		/**
		* Video matching method -  called once for each pair of videos in the list.
		* @param matchResults container for the results of matching
		* @param qdescrname input query descriptor name
		* @param rdescrname input reference descriptor name
		* @param qbitrate query bitrate (one of 16,64,256)
		* @param rbitrate reference bitrate (one of 16,64,256)
		* @return the matching score (normalized in the [0..1] range)
		*/
		virtual double match(MatchData & matchResults, const std::string & qdescrname, const std::string & rdescrname, int qbitrate, int rbitrate);

		/**
		* Video indexing method - builds a DB of reference video descriptors.
		* @param cdva_descriptor the descriptor to add to the DB
		* @param relativepathname the relative pathname of the video file to be used as unique identifier
		*/
		virtual void makeindex(const std::string & cdva_descriptor, const std::string & relativepathname);

		/**
		* Video retrieval method - returns a list of reference videos matching the query video.
		* @param retrievalResults the output vector containing an ordered list of matching reference videos
		* @param qdescrname the video query descriptor
		* @param qbitrate query bitrate (one of 16,64,256)
		*/
		virtual void retrieve(std::vector<MatchData> & retrievalResults, const std::string & qdescrname, int qbitrate);

		/**
		 * Save database information.
		 */
		void commitDB();

		/**
		* de-initialization method - called once at the end of processing.
		*/
		virtual void close();

	}; 	// end class CDVAImpl


}  // end of namespace mpeg7cdva
