/*
The copyright in this software is being made available under this MPEG Reference Software Copyright License. This software may be subject to other third party and contributor rights, including patent rights, and no such rights are granted under this license.

Copyright (c) 2016-2017, Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
*	Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
*	Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
*	Neither the names of Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms nor the names of their contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "CompressedFeatureList.h"

#include "Match.h"
#include <iostream>

using namespace mpeg7cdvs;

#define HALF_PI (1.570796326794897)

namespace mpeg7cdva {


LeanCoordList::LeanCoordList() {
	nFeatures = 0;
	xCoord = 0;
	yCoord = 0;
}

LeanCoordList::LeanCoordList(int size) {
	nFeatures = size;

	xCoord = new unsigned int[nFeatures];
	yCoord = new unsigned int[nFeatures];

	
}

LeanCoordList::LeanCoordList(FeatureList& fl){
	nFeatures = fl.nFeatures();
	
	xCoord = new unsigned int[nFeatures];
	yCoord = new unsigned int[nFeatures];

	for (int i = 0; i < nFeatures; i++) {
		xCoord[i] = fl.features[i].x;
		yCoord[i] = fl.features[i].y;
	}

}

LeanCoordList::LeanCoordList(const LeanCoordList& other) {

	nFeatures = other.nFeatures;
	xCoord = new unsigned int[nFeatures];
	yCoord = new unsigned int[nFeatures];

	memcpy(xCoord, other.xCoord, nFeatures*sizeof(unsigned int));
	memcpy(yCoord, other.yCoord, nFeatures*sizeof(unsigned int));
}

LeanCoordList & LeanCoordList::operator= (const LeanCoordList & other) {
	if (nFeatures>0) {
		delete[] xCoord;
		delete[] yCoord;
	}

	nFeatures = other.nFeatures;
	xCoord = new unsigned int[nFeatures];
	yCoord = new unsigned int[nFeatures];

	memcpy(xCoord, other.xCoord, nFeatures*sizeof(unsigned int));
	memcpy(yCoord, other.yCoord, nFeatures*sizeof(unsigned int));

	return *this;
}

mpeg7cdvs::FeatureList& LeanCoordList::getFeatureList() {
	mpeg7cdvs::FeatureList fl;


	for (int i = 0; i < nFeatures; i++) {
		Feature f;
		f.x = xCoord[i];
		f.y = yCoord[i];
		fl.addFeature(f);
	}

	return fl;
}

void LeanCoordList::getFeatureList(mpeg7cdvs::FeatureList& fl) {
	fl.features.resize(nFeatures);

	for (int i = 0; i < nFeatures; i++) {
		fl.features[i].x = xCoord[i];
		fl.features[i].y = yCoord[i];
	
	}
}

LeanCoordList::~LeanCoordList() {
	delete[] xCoord;
	delete[] yCoord;
}


CompressedFeatureList::CompressedFeatureList()
{
	originalWidth = 0;
	originalHeight = 0;
	imageHeight = 0;
	imageWidth = 0;
}



CompressedFeatureList::CompressedFeatureList(const CompressedFeatureList& a, bool relevantOnly) {
	originalWidth = a.originalWidth;
	originalHeight = a.originalHeight;
	imageHeight = a.imageHeight;
	imageWidth = a.imageWidth;

	int numFeatures = a.numFeatures;
	int nrel = 0;;
	for (int i = 0; i<a.relevances.size(); i++) if (a.relevances[i]>0) nrel++;

	if (relevantOnly) numFeatures = nrel;

	relevances.resize(numFeatures);

	if (a.numFeatures > 0)	// really copy only if not empty
	{
		imagefile = a.imagefile;
		allocate(numFeatures, a.nDescLength);

		if (a.numFeatures == numFeatures) {

			memcpy(Xcoord, a.Xcoord, numFeatures * sizeof(unsigned short));
			memcpy(Ycoord, a.Ycoord, numFeatures * sizeof(unsigned short));
			if (features != NULL)
			{
				memcpy(features, a.features, numFeatures * nDescLength);
			}
			for (int i = 0; i < a.numFeatures; i++) relevances[i] = a.relevances[i];
		}
		else {
			int j = 0;
			for (int i = 0; i < a.numFeatures; i++) {
				if (a.relevances[i]>0) {
					Xcoord[j] = a.Xcoord[i];
					Ycoord[j] = a.Ycoord[i];
					features[j] = a.features[i];
					relevances[j] = a.relevances[i];
					j++;
				}
			}
		}
	}

	
}



CompressedFeatureList::CompressedFeatureList(const mpeg7cdvs::FeatureList & a, bool relevantOnly)
{
	originalWidth = a.originalWidth;
	originalHeight = a.originalHeight;
	imageHeight = a.imageHeight;
	imageWidth = a.imageWidth;

	int nBytes = a.qdescr_size >> 2;
	int nFeatures = relevantOnly ? a.getRelevantPoints() : a.nFeatures();

	allocate(nFeatures, nBytes);
	int k = 0;

	for (std::vector<Feature>::const_iterator f = a.features.begin(); f<a.features.end(); ++f)
	{
		if ((relevantOnly) && (f->relevance == 0))
			continue;		// skip all features having low relevance

		Xcoord[k] = (unsigned short)(f->x + 0.5f);			// round coordinates
		Ycoord[k] = (unsigned short)(f->y + 0.5f);			// round coordinates

		// Pack descriptor elements into bytes: 4 elements per byte.
		unsigned char uc;
		int nof_elements = a.qdescr_size;
		int byte_offset = 0;
		int descr_bytes = 0;
		for (int i = 0; i < nof_elements; i++)
		{
			if (byte_offset == 0)
			{
				uc = 0;
			}
			int val = f->qdescr[i];
			if (val == 2) val = 3;
			uc |= (val & 3) << byte_offset;
			byte_offset += 2;
			if (byte_offset == 8)
			{
				features[k*nDescLength + descr_bytes] = uc;
				byte_offset = 0;
				// go to next byte;
				descr_bytes++;
			}
		}

		++k;	// go to the next compressed feature
	}
}

void CompressedFeatureList::allocate(int nFeatures, int descLen)
{
	if ((nFeatures <= 0) || (nFeatures > 16000))
		throw CdvsException("invalid parameter nFeatures passed to CompressedFeatureList::allocate");

	if ((descLen <= 0) || (descLen > 32))
		throw CdvsException("invalid parameter descLen passed to CompressedFeatureList::allocate");

	numFeatures = nFeatures;
	nDescLength = descLen;
	features = new unsigned char[nFeatures * descLen];


	// allocate array for X and Y coordinates of key points

	Xcoord = new unsigned short[nFeatures];
	Ycoord = new unsigned short[nFeatures];

	relevances.resize(numFeatures);
}

// copy and swap idiom
CompressedFeatureList & CompressedFeatureList::operator=(CompressedFeatureList a)
{
	a.swap(*this);
	relevances = a.relevances;
	return *this;
}

void CompressedFeatureList::toCDVScfl(mpeg7cdvs::CompressedFeatureList& cfl) {
	swap(cfl);
}


CompressedFeatureList & CompressedFeatureList::operator=(mpeg7cdvs::CompressedFeatureList a)
{
	a.swap(*this);
	relevances.resize(numFeatures);
	return *this;
}


	int CompressedFeatureList::matchDescriptors_oneWay(PointPairs &pairs, const CompressedFeatureList & otherFeatureList, float ratioThreshold, 
		DescriptorTimeMap& thisDtm, DescriptorTimeMap& otherDtm, CoordList& thisFl, CoordList& otherFl, int thisIdx, int otherIdx, int** distanceMatrix) const
{
	pairs.nMatched = 0;

	if ((numFeatures > 1) && (otherFeatureList.numFeatures > 1))
	{
		ratioThreshold = ratioThreshold*ratioThreshold;

		int distance;
		int minDistance, secondMinDistance;
		int minDistanceInd;

		Match match;
		std::vector<Match> matches;
		std::vector<int> featureMatches;

		// Check sizes of the descriptors.
		int matched_bytes = std::min(4*nDescLength, 4*otherFeatureList.nDescLength);

		//// Select the two nearest descriptors
		for (int featureInd = 0; featureInd < numFeatures; ++featureInd)
		{
			// skip features not used in this frame
			if (!thisDtm.get(featureInd, thisIdx)) continue;



			// Find the two nearest descriptors contained in otherFeatureList and the relative distances between f

			minDistance = 65536;
			secondMinDistance = 65536;
			minDistanceInd = 0;

			for (int otherFeatureInd = 0; otherFeatureInd < otherFeatureList.numFeatures; ++otherFeatureInd)
			{


				// skip features not used in this frame
				if (!otherDtm.get(otherFeatureInd, otherIdx)) continue;

				if (distanceMatrix[featureInd][otherFeatureInd]>=0) distance = distanceMatrix[featureInd][otherFeatureInd];
				else {
					distance = getDistance(features + featureInd*nDescLength, otherFeatureList.features + otherFeatureInd*otherFeatureList.nDescLength, matched_bytes);
					distanceMatrix[featureInd][otherFeatureInd] = distance;
					
				}


				if (distance<minDistance)
				{
					secondMinDistance = minDistance;
					minDistance = distance;
					minDistanceInd = otherFeatureInd;
				}
				else
				{
					if (distance<secondMinDistance)
					{
						secondMinDistance = distance;
					}
				}
			}	// end for otherFeatureInd

			// If the ratio test is passed the indices of the features are saved
			if ((minDistance <= ratioThreshold*secondMinDistance) && (secondMinDistance > 0))
			{
				match.featureInd = featureInd;
				match.otherFeatureInd = minDistanceInd;
				match.weight = cos(HALF_PI * sqrt((double)minDistance / (double)secondMinDistance));
				matches.push_back(match);
			}
		}	// end for featureInd

		// Creation of a list on biunique matchings deleting possible repetitions in otherFeatureInd
		if (matches.size()>0)
		{
			// Ordering matches in accordance to otherFeatureInd
			std::sort(matches.begin(), matches.end(), Match::sortMatchByWeight);
			int lastFeatureIndex = -1;

			// Look for repetitions
			for (std::vector<Match>::const_iterator m = matches.begin(); m<matches.end(); ++m)
			{
				if (lastFeatureIndex != m->otherFeatureInd)
				{
					lastFeatureIndex = m->otherFeatureInd;
					for (int kk = 0; kk < thisDtm.getCount(m->featureInd, thisIdx);kk++) {
						for (int ll = 0; ll < otherDtm.getCount(m->otherFeatureInd, otherIdx); ll++) {

							int thisCoordF = thisDtm.getFeatureStartIndex(m->featureInd, thisIdx) + kk;
							int otherCoordF = otherDtm.getFeatureStartIndex(m->otherFeatureInd, otherIdx) + ll;

							pairs.addPair((unsigned short)thisFl.getX(thisCoordF) + 0.5,
								(unsigned short)otherFl.getX(otherCoordF) + 0.5,
								(unsigned short)thisFl.getY(thisCoordF) + 0.5,
								(unsigned short)otherFl.getY(otherCoordF) + 0.5,
								m->weight);
						}

					}

				}
			}
		}
	}

	return pairs.nMatched;
}


int CompressedFeatureList::matchDescriptors_twoWay(PointPairs &pairs, const CompressedFeatureList &otherFeatureList, float ratioThreshold, 
	DescriptorTimeMap& thisDtm, DescriptorTimeMap& otherDtm, CoordList& thisFl, CoordList& otherFl, int thisIdx, int otherIdx, int** distanceMatrix) const
{
	pairs.nMatched = 0;
	std::vector<Match> matches;
	Match match;
	std::vector<Match> matches1;
	std::vector<Match> matches1clean;
	std::vector<Match> matches2;
	std::vector<Match> matches2clean;

	if (numFeatures>1 && otherFeatureList.numFeatures>1)
	{
		// There are at least two features in each featureList

		// We work with the square of the distances to optimize the computational time
		// This requires to raise ratioThreshold to a power of two
		ratioThreshold *= ratioThreshold;

		int distance;
		int minDistance, secondMinDistance;
		int minDistanceInd;


		std::vector<int> featureMatches;

		// Check sizes of the descriptors.
		int matched_bytes = std::min(nDescLength, otherFeatureList.nDescLength);

		// Select the two nearest descriptors

		for (int featureInd = 0; featureInd < numFeatures; ++featureInd)
		{

			// skip features not used in this frame
			if (!thisDtm.get(featureInd, thisIdx)) continue;

			// Find the two nearest descriptors contained in otherFeatureList and the relative distances between f
			minDistance = 65536;
			secondMinDistance = 65536;
			minDistanceInd = 0;

			for (int otherFeatureInd = 0; otherFeatureInd < otherFeatureList.numFeatures; ++otherFeatureInd)
			{

				// skip features not used in this frame
				if (!otherDtm.get(otherFeatureInd, otherIdx)) continue;
				
				if (distanceMatrix[featureInd][otherFeatureInd]>=0) distance = distanceMatrix[featureInd][otherFeatureInd];
				else {
					distance = getDistance(features + featureInd*nDescLength, otherFeatureList.features + otherFeatureInd*otherFeatureList.nDescLength, matched_bytes);
					distanceMatrix[featureInd][otherFeatureInd] = distance;
				}


				if (distance<minDistance)
				{
					secondMinDistance = minDistance;
					minDistance = distance;
					minDistanceInd = otherFeatureInd;
				}
				else
				{
					if (distance<secondMinDistance)
					{
						secondMinDistance = distance;
					}
				}
			}	// end for otherFeatureInd

			// If the ratio test is passed the indices of the features are saved
			if ((minDistance <= ratioThreshold*secondMinDistance) && (secondMinDistance>0))
			{
				match.featureInd = featureInd;
				match.otherFeatureInd = minDistanceInd;
				match.weight = cos(HALF_PI * sqrt((float)minDistance / ((float)secondMinDistance)));
				matches1.push_back(match);
			}
		}	// end for featureind

		// Creation of a list on biunique matchings deleting possible repetitions in otherFeatureInd
		if (matches1.size()>0)
		{
			// Ordering matches in accordance to otherFeatureInd
			std::sort(matches1.begin(), matches1.end(), Match::sortMatchByWeight);
			int lastFeatureIndex = -1;

			// Look for repetitions
			for (std::vector<Match>::const_iterator m = matches1.begin(); m<matches1.end(); ++m)
			{
				if (lastFeatureIndex != m->otherFeatureInd)
				{
					lastFeatureIndex = m->otherFeatureInd;
					Match tempMatch;
					tempMatch.featureInd = m->featureInd;
					tempMatch.otherFeatureInd = m->otherFeatureInd;
					tempMatch.weight = m->weight;
					matches1clean.push_back(tempMatch);
				}
			}
		}
		matches1.clear();

		// Select the two nearest descriptors

		for (int otherFeatureInd = 0; otherFeatureInd < otherFeatureList.numFeatures; ++otherFeatureInd)
		{
			// skip features not used in this frame
			if (!otherDtm.get(otherFeatureInd, otherIdx)) continue;

			// Find the two nearest descriptors contained in otherFeatureList and the relative distances between f
			minDistance = 65536;
			secondMinDistance = 65536;
			minDistanceInd = 0;

			for (int featureInd = 0; featureInd < numFeatures; ++featureInd)
			{			
				
				// skip features not used in this frame
				if (!thisDtm.get(featureInd, thisIdx)) continue;
				
				distance = distanceMatrix[featureInd][otherFeatureInd];

				if (distance<minDistance)
				{
					secondMinDistance = minDistance;
					minDistance = distance;
					minDistanceInd = featureInd;
				}
				else
				{
					if (distance<secondMinDistance)
					{
						secondMinDistance = distance;
					}
				}
			}	// end for featureInd

			// If the ratio test is passed the indices of the features are saved
			if ((minDistance < ratioThreshold*secondMinDistance) && (secondMinDistance > 0))
			{
				match.featureInd = minDistanceInd;
				match.otherFeatureInd = otherFeatureInd;
				match.weight = cos(HALF_PI * sqrt((float)minDistance / ((float)secondMinDistance)));
				matches2.push_back(match);
			}
		}	// end for otherFeatureInd

		// Creations of a list on biunique matchings deleting possible repetitions in otherFeatureInd
		if (matches2.size()>0)
		{
			// Ordering matches in accordance to otherFeatureInd
			std::sort(matches2.begin(), matches2.end(), Match::sortMatchByWeight);
			int lastFeatureIndex = -1;

			// Look for repetitions
			for (std::vector<Match>::const_iterator m = matches2.begin(); m<matches2.end(); ++m)
			{
				if (lastFeatureIndex != m->featureInd)
				{
					lastFeatureIndex = m->featureInd;
					Match tempMatch;
					tempMatch.featureInd = m->featureInd;
					tempMatch.otherFeatureInd = m->otherFeatureInd;
					tempMatch.weight = m->weight;
					matches2clean.push_back(tempMatch);
				}
			}
		}
		matches2.clear();

		for (std::vector<Match>::iterator m1 = matches1clean.begin(); m1 < matches1clean.end(); ++m1)
		{
			int flag = 0;
			for (std::vector<Match>::iterator m2 = matches2clean.begin(); m2 < matches2clean.end(); ++m2)
			{
				if (m1->featureInd == m2->featureInd && m1->otherFeatureInd == m2->otherFeatureInd)
				{
					for (int kk = 0; kk < thisDtm.getCount(m1->featureInd, thisIdx); kk++) {
						for (int ll = 0; ll < otherDtm.getCount(m1->otherFeatureInd, otherIdx); ll++) {

							int thisCoordF = thisDtm.getFeatureStartIndex(m1->featureInd, thisIdx) + kk;
							int otherCoordF = otherDtm.getFeatureStartIndex(m1->otherFeatureInd, otherIdx) + ll;

							pairs.addPair((unsigned short)thisFl.getX(thisCoordF) + 0.5,
								(unsigned short)otherFl.getX(otherCoordF) + 0.5,
								(unsigned short)thisFl.getY(thisCoordF) + 0.5,
								(unsigned short)otherFl.getY(otherCoordF) + 0.5,
								(m1->weight + m2->weight)*0.5, match_2way_INTERSECTION);
							flag = 1;
							break;
						}
					}
				}
			}
			if (flag == 0)
			{
				for (int kk = 0; kk < thisDtm.getCount(m1->featureInd, thisIdx); kk++) {
					for (int ll = 0; ll < otherDtm.getCount(m1->otherFeatureInd, otherIdx); ll++) {

						int thisCoordF = thisDtm.getFeatureStartIndex(m1->featureInd, thisIdx) + kk;
						int otherCoordF = otherDtm.getFeatureStartIndex(m1->otherFeatureInd, otherIdx) + ll;

						pairs.addPair((unsigned short)thisFl.getX(thisCoordF) + 0.5,
							(unsigned short)otherFl.getX(otherCoordF) + 0.5,
							(unsigned short)thisFl.getY(thisCoordF) + 0.5,							
							(unsigned short)otherFl.getY(otherCoordF) + 0.5,
							m1->weight, match_2way_DISJOINT1);
					}
				}
			}
		}

		for (std::vector<Match>::iterator m2 = matches2clean.begin(); m2 < matches2clean.end(); ++m2)
		{
			int flag = 0;
			for (std::vector<Match>::iterator m1 = matches1clean.begin(); m1 < matches1clean.end(); ++m1)
			{
				if (m1->featureInd == m2->featureInd && m1->otherFeatureInd == m2->otherFeatureInd){ flag = 1; break; }
			}
			if (flag == 0)
			{
				for (int kk = 0; kk < thisDtm.getCount(m2->featureInd, thisIdx); kk++) {
					for (int ll = 0; ll < otherDtm.getCount(m2->otherFeatureInd, otherIdx); ll++) {

						int thisCoordF = thisDtm.getFeatureStartIndex(m2->featureInd, thisIdx) + kk;
						int otherCoordF = otherDtm.getFeatureStartIndex(m2->otherFeatureInd, otherIdx) + ll;

						pairs.addPair((unsigned short)thisFl.getX(thisCoordF) + 0.5,
							(unsigned short)otherFl.getX(otherCoordF) + 0.5,
							(unsigned short)thisFl.getY(thisCoordF) + 0.5,
							(unsigned short)otherFl.getY(otherCoordF) + 0.5,
							m2->weight, match_2way_DISJOINT2);
					}
				}
			}
		}

		double disjoin_weigh = 0;
		int disjoint_matches = 0;
		int intersection_matches = 0;
		for (int i = 0; i < pairs.nMatched; i++)
		{
			if (pairs.match_dirs[i] == match_2way_INTERSECTION)
				intersection_matches++;
			else
				disjoint_matches++;
		}
		if (disjoint_matches > intersection_matches)
		{
			disjoin_weigh = intersection_matches / (double)disjoint_matches;
			for (int i = 0; i < pairs.nMatched; i++)
			{
				if (pairs.match_dirs[i] != match_2way_INTERSECTION)
					pairs.weights[i] *= disjoin_weigh;
			}
		}


	}
	return pairs.nMatched;

}

void CompressedFeatureList::getFeatureList(mpeg7cdvs::FeatureList& fl) {

	// unpack descriptor representation
	// coordinates are not modified

	int k = 0;

	fl.qdescr_size = nDescLength << 2;

	for (std::vector<Feature>::iterator f = fl.features.begin(); f<fl.features.end(); ++f)
	{
		// Pack descriptor elements into bytes: 4 elements per byte.
		unsigned char uc;
		int nof_elements = fl.qdescr_size;
		int byte_offset = 8;
		int descr_bytes = 0;
		for (int i = 0; i < nof_elements; i++)
		{
			if (byte_offset == 8) {
				uc = features[k*nDescLength + descr_bytes];
				byte_offset = 0;
				// go to next byte;
				descr_bytes++;
			}

			int val = (uc & (3 << byte_offset)) >> byte_offset;
			if (val == 3) val = 2;
			f->qdescr[i] = val;
			byte_offset += 2;

		}

		/*
		
				for (int i = 0; i < nof_elements; i++)
		{
			if (byte_offset == 0)
			{
				uc = 0;
			}
			int val = f->qdescr[i];
			if (val == 2) val = 3;
			uc |= (val & 3) << byte_offset;
			byte_offset += 2;
			if (byte_offset == 8)
			{
				features[k*nDescLength + descr_bytes] = uc;
				byte_offset = 0;
				// go to next byte;
				descr_bytes++;
			}
		}
		*/


		++k;	// go to the next compressed feature
	}

}


}