/*
The copyright in this software is being made available under this MPEG Reference Software Copyright License. This software may be subject to other third party and contributor rights, including patent rights, and no such rights are granted under this license.

Copyright (c) 2016-2017, Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
*	Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
*	Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
*	Neither the names of Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms nor the names of their contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#pragma once

#include "BitOutputStream.h"
#include "BitInputStream.h"
#include "FeatureList.h"
#include "Parameters.h"
#include "RefFeature.h"
#include "DescriptorTimeMap.h"

using namespace mpeg7cdvs;

namespace mpeg7cdva
{

/**
 * @class CsscCoordinateCoding
 * Class that converts the coordinates of all descriptors of an image into a bitstream, and vice versa.
 * @date 2012
 */
class CoordinateCoding
{
public:
	static const int	SUM_HIST_COUNT_SIZE = 64;										///< Histogram count context lenght
	static const int	CONTEXT_RANGE = 5;												///< Sum-based context range					
	static const int 	MAXIMUM_SUM_CONTEXT = (2*CONTEXT_RANGE*CONTEXT_RANGE + CONTEXT_RANGE) ;	///< Maximum value of sum context
	
	/**
	 * @struct CircularSumContext
	 * Basic structure for Cssc
	 */
	struct CircularSumContext {
		long	 vCount[SUM_HIST_COUNT_SIZE];		///< Histogram Count Arithmetic Coding model initialization data
		long	 vInitialMap[2];					///< Circular scan map Arithmetic Coding model initialization data (initial)
		long	 vMap[MAXIMUM_SUM_CONTEXT+1][2];	///< Circular scan map Arithmetic Coding model initialization data (following)
	} ;

private:

	static const int CONTEXT_TABLE[12][6];
	static const int HIST_COUNT_TABLE[12][3];

	CoordinateCoding(const CoordinateCoding& p);									// copy constructor - set private to avoid its usage.
	CoordinateCoding &operator=(const CoordinateCoding&);							// assignment operator - set private to avoid its usage.
	static int compareArray(int * a, int * b, int size);									// compare two int arrays
	static int compareArray(int * a, int * b, int iWidth, int iHeight, bool bTranspose);	// compare two int arrays
	static int compareArray(bool * a, bool * b, int size);									// compare two bool arrays

	// Input parameters
	int 	m_debugLevel;		// If (debugLevel > 0) some debugging info will be printed
	int 	m_blockWidth;		// Size of the grid used to build the histogram map
	int		m_queryLength;		// To identify operating point

	// Histogram count data structure
	int* 	m_hCount; 			// Pointer to the histogram count 
	int 	m_hCountSize;		// Histogram count size

	// Histogram map data structure
	int* 	m_hMap;				// Pointer to histogram map
	int 	m_hMapSizeX;		// X Dimension of histogram map
	int 	m_hMapSizeY;		// Y Dimension of histogram map
	bool 	m_bIsTransposed;    // Indicating whether the circular scanned matrix needs to be transposed

	// For context entropy coding
	bool 	m_bTrainingMode;
	CircularSumContext	m_csc;	

	bool m_readSize;

	// Utility functions
	void 	TransposeMatrix( int* &pMatrix, int &iRows, int &iCols);

	bool readOK;

public:	

	/**
	 * Constructor using the given parameters to set the CsscCoordinateCoding behaviour.
	 * @param param the set of parameters to initialize this object.
	 */
	CoordinateCoding(const Parameters & param);

	virtual ~CoordinateCoding();							// destructor

	/**
	 * Convert the stored information into a binary stream.
	 * @param writer the bitstream writer object.
	 */
	void toBinary(BitOutputStream & writer, bool writeSize = true);

	/**
	 * Convert a binary stream into the stored information.
	 * @param reader the bitstream reader object.
	 */
	void fromBinary(BitInputStream & reader);
	 
	/**
	* Convert a binary stream into the stored information.
	* In this case the buffer does not contain histo X and Y size, but they are specified as parameters.
	* @param reader the bitstream reader object.
	* @param
	*/
	void fromBinary(BitInputStream & reader, unsigned int hMapSizeX, unsigned int hMapSizeY);

	/**
	 * Compare this instance with another one. 
	 * @param other the other instance.
	 * @return 0 if equal, or the number of different values if different.
	 */
	int compare(const CoordinateCoding & other);

	/**
	 * Export the value of the histogram count and size.
	 * @param histogramCountSize size of histogram count
	 * @param histogramMapSizeX	size of histogram map (X)
	 * @param histogramMapSizeY size of histogram map (Y)
	 */
	void exportVars(unsigned int & histogramCountSize, unsigned int & histogramMapSizeX, unsigned int & histogramMapSizeY) const;


	/**
	 * Generation of new matrix representation based on circular scanning. 
	 * @param featurelist list of keypoints.
	 * @param numPoints the number of features to encode (only the first numPoint features in featurelist will be encoded)
	 */
	void 	generateHistogramMap(FeatureList & featurelist, int numPoints);

	/**
	* Generation of new matrix representation based on circular scanning.
	* keeps order specified by ref feature list (using populated entries only
	* @param featurelist list of keypoints.
	* @param numPoints the number of features to encode (only the first numPoint features in featurelist will be encoded)
	*/
	void generateHistogramMap(std::vector<RefFeature> & refFeaturelist, FeatureList & featurelist, int offset, int frameIdx, DescriptorTimeMap& dtm, int numPoints, int xSize, int ySize);


	/**
	 * Reconstruction of the original histogram map starting from circular scanning representation. 
	 * @param descriptors reconstructed list of keypoints and descriptors
	 */
	void 	generateFeatureList(FeatureList & descriptors);


	// Functions for context training
	void 	StartTrainingMode();
	void 	EndTrainingMode();
	int 	AddImageSample(FeatureList& featurelist);

	// Functions for context I/O
	int     writeSeparateContext(char *filename);
	static int		readSeparateContext( char *filename, CircularSumContext &cCsc);
	int		readSeparateContext( char *filename);

	//////////////

	void setVerbose() { m_debugLevel = 2; }

	/** Set object to be the differnce of the current histogram and the one passed as argument 
	  * same image size is assumed
	  * call after generateHistogramMap2
	  */
	void difference(CoordinateCoding& other);

	void 	generateHistogramMap2(FeatureList & featurelist, int numPoints);

	void setIsOK(bool ok) { readOK = ok;  }

};

} 	// end of namespace
