/*
The copyright in this software is being made available under this MPEG Reference Software Copyright License. This software may be subject to other third party and contributor rights, including patent rights, and no such rights are granted under this license.

Copyright (c) 2016-2017, Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
*	Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
*	Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
*	Neither the names of Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms nor the names of their contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "DescriptorTimeMap.h"

#include <stdio.h>
#include <string.h>
#include <memory>

namespace mpeg7cdva {

DescriptorTimeMap::DescriptorTimeMap() :
	m_nDescr(0), m_nAbsDescr(0), m_nFrames(0)

{
	m_nbits = 0;

	m_buffer = 0;

	bufsz = 0;
}

DescriptorTimeMap::DescriptorTimeMap(int nDescr, int nAbsDescr, int nFrames) : 
	m_nDescr(nDescr), m_nAbsDescr(nAbsDescr), m_nFrames(nFrames)

{
	m_nbits = m_nDescr*m_nFrames*4;

	bufsz = (int)ceil(m_nbits / 8.0);

	m_buffer = new char[bufsz];

	memset(m_buffer, 0, bufsz);

}


DescriptorTimeMap::DescriptorTimeMap(int nDescr, int nAbsDescr, int nFrames, char* buf) 

{
	m_nDescr = nDescr;
	m_nAbsDescr = nAbsDescr;
	m_nFrames = nFrames;

	m_nbits = m_nDescr*m_nFrames*4;

	bufsz = (int)ceil(m_nbits / 8.0);

	m_buffer = new char[bufsz];

	memcpy(m_buffer, buf, bufsz);

	descIndex.resize(m_nFrames);
}

DescriptorTimeMap::DescriptorTimeMap(DescriptorTimeMap& other) :
m_nDescr(other.m_nDescr), m_nAbsDescr(other.m_nAbsDescr), m_nFrames(other.m_nFrames)
{
	m_nbits = m_nDescr*m_nFrames*4;

	bufsz = (int)ceil(m_nbits / 8.0);

	m_buffer = new char[bufsz];

	memcpy(m_buffer, other.m_buffer, bufsz);

	descIndex.resize(m_nFrames);
}

DescriptorTimeMap & DescriptorTimeMap::operator= (const DescriptorTimeMap & other) {
	m_nDescr = other.m_nDescr;
	m_nAbsDescr = other.m_nAbsDescr;
	m_nFrames = other.m_nFrames;

	m_nbits = m_nDescr*m_nFrames*4;

	bufsz = (int)ceil(m_nbits / 8.0);

	m_buffer = new char[bufsz];

	memcpy(m_buffer, other.m_buffer, bufsz);

	descIndex.resize(m_nFrames);

	return *this;
}

void DescriptorTimeMap::init(int nDescr, int nAbsDescr, int nFrames) {
	if (m_buffer) delete[] m_buffer;

	m_nDescr = nDescr;
	m_nAbsDescr = nAbsDescr;
	m_nFrames = nFrames;

	m_nbits = m_nDescr*m_nFrames*4;

	bufsz = (int)ceil(m_nbits / 8.0);

	m_buffer = new char[bufsz];

	descIndex.resize(m_nFrames);

	memset(m_buffer, 0, bufsz);
}

void DescriptorTimeMap::addDescIndex(std::vector<unsigned short>& di, int frame) {
	for (int i = 0; i < di.size(); i++) {
		add(di[i], frame);
	}
}

DescriptorTimeMap::~DescriptorTimeMap() {
	if (m_buffer) delete[] m_buffer;
}


void DescriptorTimeMap::writeDescIndex(unsigned char* buffer, bool compressed, int& bufSz) const
{
	unsigned int outIdx = 2;
	
	if (!compressed) {

		for (int i = 0; i < descIndex.size(); i++) {
			for (int j = 0; j < descIndex[i].size(); j++) {
				unsigned short* usBuf = (unsigned short*)buffer;
				unsigned short s = descIndex[i][j];
				usBuf[outIdx] = s;
				outIdx++;
			}
		}
	}
	else {
		unsigned char ub = 0;
		int sh = 0;
		for (int i = 0; i < descIndex.size(); i++) {
			unsigned int featureOffs = descIndex[i][0];
			bool foInit = false;
			for (int j = 0; j < descIndex[i].size(); j++) {
				if (descIndex[i][j] == featureOffs) {
					sh++;
					if (sh == 8) {
						buffer[outIdx] = ub;
						outIdx++;
						sh = 0;
						ub = 0;
					}
					// write first index of features in this frame
					if (!foInit) {
						foInit = true;

						featureOffs = descIndex[i][j];

						unsigned short uc = descIndex[i][j];
						uc = uc & 0x3FFF; // set top bits 0
						for (int k = 13; k >= 0; k--) {
							ub = ub + (((uc & (1 << k)) >> k) << sh);
							sh++;
							if (sh == 8) {
								buffer[outIdx] = ub;
								outIdx++;
								sh = 0;
								ub = 0;
							}
						}
					}

					featureOffs++;
				}
				else {
					unsigned short uc = descIndex[i][j];
					uc = uc & 0x3FFF; // set top bits 0
					uc = uc + 0x4000; // set 1 bit to indicate abs
					for (int k = 14; k >= 0; k--) {
						ub = ub + (((uc & (1 << k)) >> k) << sh);
						sh++;
						if (sh == 8) {
							buffer[outIdx] = ub;
							outIdx++;
							sh = 0;
							ub = 0;
						}
					}

				}
			}
		}
		if (sh > 0) {
			buffer[outIdx] = ub;
			outIdx++;
		}

	}

	bufSz = outIdx;

	unsigned short* bufUS = (unsigned short*)buffer;
	bufUS[0] = bufSz-2;


	// DEBUG read back
	/*
	std::vector< std::vector<unsigned short> > descIndexCp = descIndex;
	((DescriptorTimeMap*)this)->descIndex.clear();

	std::vector<unsigned int> nLocalDesc;
	for (int i = 0; i < descIndexCp.size(); i++) {
		nLocalDesc.push_back(descIndexCp[i].size());

	}

	((DescriptorTimeMap*)this)->descIndex.resize(descIndexCp.size());

	((DescriptorTimeMap*)this)->parseDescIndex(buffer, nLocalDesc, true);

	for (int i = 0; i < descIndex.size(); i++) {
		for (int j = 0; j < descIndex[i].size(); j++) {
			if (descIndex[i][j] != descIndexCp[i][j])
			{
				bool err = true;
			}
		}
	}
	*/

}


void DescriptorTimeMap::parseDescIndex(unsigned char* buffer, std::vector<unsigned int>& nLocalDesc, bool compressed) {
	unsigned int inIdx = 0;

	if (!compressed) {
		for (int i = 0; i < nLocalDesc.size(); i++) {
			std::vector<unsigned short> diI;
			for (int j = 0; j < nLocalDesc[i]; j++) {
				unsigned short* usBuf = (unsigned short*)buffer;
				diI.push_back(usBuf[inIdx]);
				inIdx++;
			}
			addDescIndex(diI, i);
		}
	}
	else {
		unsigned char ub;
		int sh = 0;

		ub = buffer[inIdx];

		for (int i = 0; i < nLocalDesc.size(); i++) {
			std::vector<unsigned short> diI;
			int featureOffs = -1;
			for (int j = 0; j < nLocalDesc[i]; j++) {
				if ((ub & (1 << sh))>0) {
					sh++;
					if (sh == 8) {
						inIdx++;
						ub = buffer[inIdx];
						sh = 0;
					}

					unsigned short uc = 0;
					for (int k = 13; k >= 0; k--) {
						uc = uc + (((ub & (1 << sh)) >> sh) << k);
						sh++;
						if (sh == 8) {
							inIdx++;
							ub = buffer[inIdx];
							sh = 0;
						}
					}

					diI.push_back(uc);

				}
				else {
					sh++;
					if (sh == 8) {
						inIdx++;
						ub = buffer[inIdx];
						sh = 0;
					}

					// for first one read start
					if (featureOffs == -1) {

						unsigned short uc = 0;
						for (int k = 13; k >= 0; k--) {
							uc = uc + (((ub & (1 << sh)) >> sh) << k);
							sh++;
							if (sh == 8) {
								inIdx++;
								ub = buffer[inIdx];
								sh = 0;
							}
						}
						featureOffs = uc;

					}
					diI.push_back(featureOffs);

					featureOffs++;
				}

			}
			addDescIndex(diI, i);
		}

	}
}

void DescriptorTimeMap::clear() {

	delete[] m_buffer;

	m_nbits = 0;

	m_buffer = 0;

	bufsz = 0;

	descIndex.clear();
}



}